"""Chat with a KB."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests
import requests_cache
from bs4 import BeautifulSoup
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class ReusableDataWrapper(BaseWrapper):
    """
    A wrapper over a reusabledata.org yaml files.

    Each yaml file is treated as a distinct dataset object;
    the structure is not altered, other than adding a ``license_text``
    field that is obtained by querying the license_link URL.

    This text can be used for evaluating automated metadata extraction.

    This is a static wrapper: it does not allow search,
    it is expected that all objects from the source are ingested.
    """

    name: ClassVar[str] = "reusabledata"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "Dataset"

    glob: bool = False

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        session = requests_cache.CachedSession("reusabledata")

        path = self.source_locator or "https://reusabledata.org/data.json"
        objs = requests.get(path).json()
        for obj in objs:
            obj_id = obj["id"]
            license_link = obj.get("license-link", None)
            if license_link:
                if license_link in [
                    "TODO",
                    "https://",
                    "inconsistent",
                    "https://civic.genome.wustl.edu/about",
                    "http://www.supfam.org/about",
                    "ftp://ftp.nextprot.org/pub/README",
                    "ftp://ftp.jcvi.org/pub/data/TIGRFAMs/COPYRIGHT",
                ]:
                    logger.warning(f"base link {license_link} for {obj_id}")
                    continue
                if license_link.startswith("ftp://"):
                    license_link = license_link.replace("ftp://", "http://")
                logger.info(f"{obj_id} license url: {license_link}")
                response = session.get(license_link)
                if not response.ok:
                    logger.warning(f"bad link {license_link} for {obj_id}")
                    continue
                data = response.text
                soup = BeautifulSoup(data, "html.parser")
                license_text = soup.get_text()
                obj["license_text"] = license_text
            yield obj



from typing import Any, Iterator, List, Tuple, Union

import yaml

from curategpt.evaluation.evaluation_datamodel import (
    AggregationMethod,
    ClassificationMetrics,
    ClassificationOutcome,
)


def calculate_metrics(
    outcomes: List[Union[ClassificationOutcome, Tuple[ClassificationOutcome, Any]]]
) -> ClassificationMetrics:
    outcomes = [
        outcome if isinstance(outcome, ClassificationOutcome) else outcome[0]
        for outcome in outcomes
    ]
    tp = outcomes.count(ClassificationOutcome.TRUE_POSITIVE)
    tn = outcomes.count(ClassificationOutcome.TRUE_NEGATIVE)
    fp = outcomes.count(ClassificationOutcome.FALSE_POSITIVE)
    fn = outcomes.count(ClassificationOutcome.FALSE_NEGATIVE)

    # Avoid division by zero
    precision = tp / (tp + fp) if tp + fp > 0 else 0
    recall = tp / (tp + fn) if tp + fn > 0 else 0
    f1 = 2 * (precision * recall) / (precision + recall) if precision + recall > 0 else 0
    accuracy = (tp + tn) / (tp + tn + fp + fn) if tp + tn + fp + fn > 0 else 0
    specificity = tn / (tn + fp) if tn + fp > 0 else 0

    return ClassificationMetrics(
        precision=precision,
        recall=recall,
        f1_score=f1,
        accuracy=accuracy,
        specificity=specificity,
        true_positives=tp,
        true_negatives=tn,
        false_positives=fp,
        false_negatives=fn,
    )


def _normalize(obj: Any) -> str:
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, dict):
        return yaml.safe_dump(obj, sort_keys=True)
    elif isinstance(obj, list):
        return yaml.safe_dump(obj, sort_keys=True)
    else:
        return str(obj)


def evaluate_predictions(obj1: Any, obj2: Any) -> Iterator[Tuple[ClassificationOutcome, str]]:
    """
    Evaluate a prediction compared to an expected value.

    Where the prediction and the expected value are lists, the results are each
    true positive, true negative, false positive.

    Where the prediction and the expected value are scalars, these are treated as if
    they are lists, thus a correct prediction is a true positive, and no false positives
    or negatives; an incorrect prediction is a false positive and a false negative.

    :param obj1:
    :param obj2:
    :return:
    """
    if isinstance(obj1, list) and isinstance(obj2, list):
        set1 = {_normalize(obj) for obj in obj1}
        set2 = {_normalize(obj) for obj in obj2}
        for x in set1.union(set2):
            if x not in set1:
                yield ClassificationOutcome.FALSE_NEGATIVE, f"{x} in {set2}"
            elif x not in set2:
                yield ClassificationOutcome.FALSE_POSITIVE, f"{x} in {set1}"
            else:
                yield ClassificationOutcome.TRUE_POSITIVE, f"{x} in both"
    else:
        yield from evaluate_predictions([obj1] if obj1 else [], [obj2] if obj2 else [])


def aggregate_metrics(
    metrics_list: List[ClassificationMetrics], method: AggregationMethod = AggregationMethod.MACRO
):
    """
    Aggregate a list of metrics.

    Note that if the evaluation task is for a single labels rather than lists,
    then this is trivially just the proportion of correct predictions.

    :param metrics_list:
    :param method:
    :return:
    """
    if method == AggregationMethod.MACRO:
        return ClassificationMetrics(
            precision=sum(m.precision for m in metrics_list) / len(metrics_list),
            recall=sum(m.recall for m in metrics_list) / len(metrics_list),
            f1_score=sum(m.f1_score for m in metrics_list) / len(metrics_list),
            accuracy=sum(m.accuracy for m in metrics_list) / len(metrics_list),
            specificity=sum(m.specificity for m in metrics_list) / len(metrics_list),
            true_positives=sum(
                m.true_positives for m in metrics_list if m.true_positives is not None
            ),
            true_negatives=sum(
                m.true_negatives for m in metrics_list if m.true_negatives is not None
            ),
            false_positives=sum(
                m.false_positives for m in metrics_list if m.false_positives is not None
            ),
            false_negatives=sum(
                m.false_negatives for m in metrics_list if m.false_negatives is not None
            ),
        )
    elif method == AggregationMethod.MICRO:
        total_tp = sum(m.precision * (m.recall * (m.precision + m.f1_score)) for m in metrics_list)
        total_fp = sum(m.f1_score - m.precision * m.recall for m in metrics_list)
        total_fn = sum((1 - m.recall) * (m.precision + m.f1_score) for m in metrics_list)
        total_tn = sum(
            m.accuracy * (m.precision + m.recall + m.f1_score + 1) - total_tp - total_fp - total_fn
            for m in metrics_list
        )

        precision = total_tp / (total_tp + total_fp)
        recall = total_tp / (total_tp + total_fn)
        f1_score = 2 * (precision * recall) / (precision + recall)
        accuracy = (total_tp + total_tn) / (total_tp + total_tn + total_fp + total_fn)
        specificity = total_tn / (total_tn + total_fp)

        return ClassificationMetrics(
            precision=precision,
            recall=recall,
            f1_score=f1_score,
            accuracy=accuracy,
            specificity=specificity,
        )
    elif method == AggregationMethod.WEIGHTED:
        total_weight = sum(
            m.precision + m.recall + m.f1_score + m.accuracy + m.specificity for m in metrics_list
        )
        return ClassificationMetrics(
            precision=sum(
                m.precision * (m.precision + m.recall + m.f1_score + m.accuracy + m.specificity)
                for m in metrics_list
            )
            / total_weight,
            recall=sum(
                m.recall * (m.precision + m.recall + m.f1_score + m.accuracy + m.specificity)
                for m in metrics_list
            )
            / total_weight,
            f1_score=sum(
                m.f1_score * (m.precision + m.recall + m.f1_score + m.accuracy + m.specificity)
                for m in metrics_list
            )
            / total_weight,
            accuracy=sum(
                m.accuracy * (m.precision + m.recall + m.f1_score + m.accuracy + m.specificity)
                for m in metrics_list
            )
            / total_weight,
            specificity=sum(
                m.specificity * (m.precision + m.recall + m.f1_score + m.accuracy + m.specificity)
                for m in metrics_list
            )
            / total_weight,
        )
    else:
        raise ValueError("Invalid aggregation method")



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TextIO

from curategpt.agents.base_agent import BaseAgent
from curategpt.evaluation.evaluation_datamodel import ClassificationMetrics


@dataclass
class BaseEvaluator(ABC):
    """Base class for evaluators."""

    agent: BaseAgent = None

    def evaluate(
        self, test_collection: str, num_tests=10000, report_file: TextIO = None, **kwargs
    ) -> ClassificationMetrics:
        """
        Evaluate the agent on a test collection.

        :param test_collection:
        :param num_tests:
        :param report_file:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def evaluate_object(self, obj, **kwargs) -> ClassificationMetrics:
        """
        Evaluate the agent on a single object.

        :param obj:
        :param kwargs:
        :return:
        """
        raise NotImplementedError



"""Wrapper for LinkML Schema documents."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

from linkml_runtime import SchemaView
from linkml_runtime.dumpers import json_dumper

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class LinkMLSchemarapper(BaseWrapper):
    """
    A wrapper over linkml schema documents.

    Each object is a schema element (class, slot, type, enum)
    """

    name: ClassVar[str] = "linkml_schema"

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        path = self.source_locator

        sv = SchemaView(str(path))
        yield from self.from_schemaview(sv)

    def from_schemaview(self, sv: SchemaView) -> Iterator[Dict]:
        schema = sv.materialize_derived_schema()
        elt_types = ["classes", "slots", "types", "enums"]
        for elt_type in elt_types:
            for elt in getattr(schema, elt_type, {}).values():
                obj = json_dumper.to_dict(elt)
                obj["@type"] = type(elt).__name__
                yield obj



"""Command line interface for curategpt."""

import csv
import gzip
import json
import logging
import sys
import tempfile
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import click
import jsonpatch
import pandas as pd
import requests
import yaml
from click_default_group import DefaultGroup
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.yamlutils import YAMLRoot
from llm import UnknownModelError, get_model, get_plugins
from llm.cli import load_conversation
from oaklib import get_adapter
from pydantic import BaseModel

from curategpt import ChromaDBAdapter, __version__
from curategpt.agents.bootstrap_agent import BootstrapAgent, KnowledgeBaseSpecification
from curategpt.agents.chat_agent import ChatAgent, ChatResponse
from curategpt.agents.concept_recognition_agent import AnnotationMethod, ConceptRecognitionAgent
from curategpt.agents.dase_agent import DatabaseAugmentedStructuredExtraction
from curategpt.agents.dragon_agent import DragonAgent
from curategpt.agents.evidence_agent import EvidenceAgent
from curategpt.agents.huggingface_agent import HuggingFaceAgent
from curategpt.agents.summarization_agent import SummarizationAgent
from curategpt.evaluation.dae_evaluator import DatabaseAugmentedCompletionEvaluator
from curategpt.evaluation.evaluation_datamodel import StratifiedCollection, Task
from curategpt.evaluation.runner import run_task
from curategpt.evaluation.splitter import stratify_collection
from curategpt.extract import AnnotatedObject
from curategpt.extract.basic_extractor import BasicExtractor
from curategpt.store import get_store
from curategpt.store.schema_proxy import SchemaProxy
from curategpt.utils.vectordb_operations import match_collections
from curategpt.wrappers import BaseWrapper, get_wrapper
from curategpt.wrappers.literature.pubmed_wrapper import PubmedWrapper
from curategpt.wrappers.ontology import OntologyWrapper

__all__ = [
    "main",
]


def dump(
    obj: Union[str, AnnotatedObject, Dict],
    format="yaml",
    old_object: Optional[Dict] = None,
    primary_key: Optional[str] = None,
) -> None:
    """
    Dump an object to stdout.

    :param obj:
    :param format:
    :param old_object: (when format=="patch")
    :param primary_key: (when format=="patch")
    :return:
    """
    if isinstance(obj, str):
        print(obj)
        return
    if isinstance(obj, AnnotatedObject):
        obj = obj.object
    if isinstance(obj, BaseModel):
        obj = obj.dict()
    if isinstance(obj, YAMLRoot):
        obj = json_dumper.to_dict(obj)
    if format is None or format == "yaml":
        set = yaml.dump(obj, sort_keys=False)
    elif format == "json":
        set = json.dumps(obj, indent=2)
    elif format == "blob":
        set = list(obj.values())[0]
    elif format == "patch":
        patch = jsonpatch.make_patch(old_object, obj)
        patch = jsonpatch.make_patch(old_object, obj)
        patch = json.loads(patch.to_string())
        if primary_key:
            pk_val = obj[primary_key]
            patch = {pk_val: patch}
        set = yaml.dump(patch, sort_keys=False)
    else:
        raise ValueError(f"Unknown format {format}")
    print(set)


# logger = logging.getLogger(__name__)

path_option = click.option("-p", "--path", help="Path to a file or directory for database.")
database_type_option = click.option(
    "-D",
    "--database-type",
    default="chromadb",
    show_default=True,
    help="Adapter to use for database, e.g. chromadb.",
)
docstore_database_type_option = click.option(
    "--docstore_database_type",
    default="chromadb",
    show_default=True,
    help="Docstore database type.",
)

model_option = click.option(
    "-m", "--model", help="Model to use for generation or embedding, e.g. gpt-4."
)
extract_format_option = click.option(
    "--extract-format",
    "-X",
    default="json",
    show_default=True,
    help="Format to use for extraction.",
)
schema_option = click.option("-s", "--schema", help="Path to schema.")
collection_option = click.option("-c", "--collection", help="Collection within the database.")
output_format_option = click.option(
    "-t",
    "--output-format",
    type=click.Choice(["yaml", "json", "blob", "csv", "patch"]),
    default="yaml",
    show_default=True,
    help="Output format for results.",
)
relevance_factor_option = click.option(
    "--relevance-factor", type=click.FLOAT, help="Relevance factor for search."
)
generate_background_option = click.option(
    "--generate-background/--no-generate-background",
    default=False,
    show_default=True,
    help="Whether to generate background knowledge.",
)
limit_option = click.option(
    "-l", "--limit", default=10, show_default=True, help="Number of results to return."
)
replace_option = click.option(
    "--replace/--no-replace",
    default=False,
    show_default=True,
    help="replace the database before indexing.",
)
append_option = click.option(
    "--append/--no-append", default=False, show_default=True, help="Append to the database."
)
encoding_option = click.option(
    "--encoding",
    default="utf-8",
    show_default=True,
    help="Encoding for files, e.g. iso-8859-1, cp1252. Specify 'detect' to infer using chardet.",
)
object_type_option = click.option(
    "--object-type",
    default="Thing",
    show_default=True,
    help="Type of object in index.",
)
description_option = click.option(
    "--description",
    help="Description of the collection.",
)
init_with_option = click.option(
    "--init-with",
    "-I",
    help="YAML string for initialization of main wrapper object.",
)
batch_size_option = click.option(
    "--batch-size", default=None, show_default=True, type=click.INT, help="Batch size for indexing."
)


def show_chat_response(response: ChatResponse, show_references: bool = True):
    """Show a chat response."""
    print("# Response:\n")
    click.echo(response.formatted_body)
    print("\n\n# Raw:\n")
    click.echo(response.body)
    if show_references:
        print("\n# References:\n")
        for ref, ref_text in response.references.items():
            print(f"\n## {ref}\n")
            print("```yaml")
            print(ref_text)
            print("```")
        print("# Uncited:")
        for ref, ref_text in response.uncited_references.items():
            print(f"\n## {ref}\n")
            print("```yaml")
            print(ref_text)
            print("```")


@click.group(
    cls=DefaultGroup,
    default="search",
    default_if_no_args=True,
)
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.version_option(__version__)
def main(verbose: int, quiet: bool):
    """
    CLI for curategpt.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    # logger = logging.getLogger()
    logging.basicConfig()
    logger = logging.root
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)
    logger.info(f"Logger {logger.name} set to level {logger.level}")


@main.command()
@path_option
@append_option
@collection_option
@model_option
@click.option("--text-field")
@object_type_option
@description_option
@database_type_option
@click.option(
    "--view",
    "-V",
    help="View/Proxy to use for the database, e.g. bioc.",
)
@click.option("--view-settings", help="YAML settings for the view wrapper.")
@click.option(
    "--glob/--no-glob", default=False, show_default=True, help="Whether to glob the files."
)
@click.option(
    "--collect/--no-collect", default=False, show_default=True, help="Whether to collect files."
)
@click.option(
    "--select",
    help="jsonpath to use to subselect from each JSON document.",
)
@click.option("--remove-field", multiple=True, help="Field to remove recursively from each object.")
@batch_size_option
@encoding_option
@click.argument("files", nargs=-1)
def index(
    files,
    path,
    append: bool,
    text_field,
    collection,
    model,
    object_type,
    description,
    batch_size,
    glob,
    view,
    select,
    collect,
    encoding,
    remove_field,
    database_type,
    view_settings,
    **kwargs,
):
    """
    Index files.

    Indexing a json file:
        curategpt index -p duckdb/cities.duckdb -c cities -D duckdb ~/json_examples/cities.json -m all-MiniLM-L6-v2

    Indexing a folder of JSON files:

        curategpt index  -c doc files/*json

    Here each file is treated as a separate object. It is loaded into the collection called 'doc'.

    Use --glob if there are too many files to expand on the command line:

        curategpt index --glob -c doc "files/*json"

    By default no transformation is performed on the objects. However, curategpt comes
    with standard views for common formats. For example, to index a folder of HPO associations

        curategpt index --view bacdive -c bacdive strains.json

    The --select option can be used to customize the path that will be used for indexing.
    For example:

         curategpt index -c cde_ncit --select '$.DataElementQueryResults' context-*.json

    This will index the DataElementQueryResults from each file.

    """
    db = get_store(database_type, path)
    db.text_lookup = text_field
    if glob:
        files = [str(gf.absolute()) for f in files for gf in Path().glob(f) if gf.is_file()]
    if view:
        view_args = {}
        if view_settings:
            view_args = yaml.safe_load(view_settings)
            logging.info(f"View settings: {view_args}")
        wrapper = get_wrapper(view, **view_args)
        if not object_type:
            object_type = wrapper.default_object_type
        if not description:
            description = f"{object_type} objects loaded from {str(files)[0:30]}"
    else:
        wrapper = None
    if collect:
        raise NotImplementedError
    if not append:
        if collection in db.list_collection_names():
            db.remove_collection(collection)
    if model is None:
        model = "openai:"
    if not files and wrapper:
        files = ["API"]
    for file in files:
        if encoding == "detect":
            import chardet

            # Read the first num_lines of the file
            lines = []
            with open(file, "rb") as f:
                try:
                    # Attempt to read up to num_lines lines from the file
                    for _ in range(100):
                        lines.append(next(f))
                except StopIteration:
                    # Reached the end of the file before reading num_lines lines
                    pass  # This is okay; just continue with the lines read so far
            # Concatenate lines into a single bytes object
            data = b"".join(lines)
            # Detect encoding
            result = chardet.detect(data)
            encoding = result["encoding"]
        logging.debug(f"Indexing {file}")
        if wrapper:
            wrapper.source_locator = file
            objs = wrapper.objects()  # iterator
        elif file.endswith(".json"):
            objs = json.load(open(file))
        elif file.endswith(".csv"):
            with open(file, encoding=encoding) as f:
                objs = list(csv.DictReader(f))
        elif file.endswith(".tsv.gz"):
            with gzip.open(file, "rt", encoding=encoding) as f:
                objs = list(csv.DictReader(f, delimiter="\t"))
        elif file.endswith(".tsv"):
            objs = list(csv.DictReader(open(file, encoding=encoding), delimiter="\t"))
        else:
            objs = yaml.safe_load_all(open(file, encoding=encoding))
        if isinstance(objs, (dict, BaseModel)):
            objs = [objs]
        if select:
            import jsonpath_ng as jp

            path_expr = jp.parse(select)
            new_objs = []
            for obj in objs:
                for match in path_expr.find(obj):
                    logging.debug(f"Match: {match.value}")
                    if isinstance(match.value, list):
                        new_objs.extend(match.value)
                    else:
                        new_objs.append(match.value)
            objs = new_objs
        if remove_field:
            raise NotImplementedError(
                "Use yq instead, e.g. yq eval 'del(.. | .evidence?)' input.yaml"
            )
        db.insert(objs, model=model, collection=collection, batch_size=batch_size)
    db.update_collection_metadata(
        collection, model=model, object_type=object_type, description=description
    )


@main.command(name="search")
@path_option
@collection_option
@limit_option
@relevance_factor_option
@database_type_option
@click.option(
    "--show-documents/--no-show-documents",
    default=False,
    show_default=True,
    help="Whether to show documents/text (e.g. for chromadb).",
)
@click.argument("query")
def search(query, path, collection, show_documents, database_type, **kwargs):
    """Search a collection using embedding search.

    curategpt search "Statue of Liberty" -p stagedb -c cities -D chromadb
    curategpt search "Statue of Liberty" -p duckdb/cities.duckdb -c cities -D duckdb --show-documents

    """
    db = get_store(database_type, path)
    results = db.search(query, collection=collection, **kwargs)
    i = 0
    for obj, distance, _meta in results:
        i += 1
        print(f"## {i} DISTANCE: {distance}")
        print(yaml.dump(obj, sort_keys=False))
        if show_documents:
            print("```\n", obj, "\n```")


@main.command(name="all-by-all")
@path_option
@collection_option
@limit_option
@relevance_factor_option
@database_type_option
@click.option(
    "--other-collection",
    "-X",
    help="Other collection to compare against.",
)
@click.option(
    "--other-path",
    "-P",
    help="Path for other collection (defaults to main path).",
)
@click.option(
    "--threshold",
    type=click.FLOAT,
    help="Cosine smilarity threshold for matches.",
)
@click.option(
    "--ids-only/--no-ids-only",
    default=False,
    show_default=True,
    help="Whether to show only ids.",
)
@click.option(
    "--left-field",
    "-L",
    multiple=True,
    help="Field to show from left collection (can provide multiple).",
)
@click.option(
    "--right-field",
    "-R",
    multiple=True,
    help="Field to show from right collection (can provide multiple).",
)
@output_format_option
def all_by_all(
    path,
    collection,
    other_collection,
    other_path,
    threshold,
    ids_only,
    output_format,
    left_field,
    right_field,
    database_type,
    **kwargs,
):
    """Match two collections

    curategpt all-by-all -p stagedb -P stagedb -c objects_a -X objects_b -D chromadb --ids-only

    """
    db = get_store(database_type, path)
    other_db = get_store(database_type, path)
    if other_path is None:
        other_path = path
    results = match_collections(db, collection, other_collection, other_db)

    def _obj(obj: Dict, is_left=False) -> Any:
        if ids_only:
            obj = {"id": obj["id"]}
        if is_left and left_field:
            return {f"left_{k}": obj[k] for k in left_field}
        if not is_left and right_field:
            return {f"right_{k}": obj[k] for k in right_field}
        side = "left" if is_left else "right"
        obj = {f"{side}_{k}": v for k, v in obj.items()}
        return obj

    i = 0
    for obj1, obj2, sim in results:
        if threshold and sim < threshold:
            continue
        i += 1
        obj1 = _obj(obj1, is_left=True)
        obj2 = _obj(obj2, is_left=False)
        row = {**obj1, **obj2, "similarity": sim}
        if output_format == "csv":
            if i == 1:
                fieldnames = list(row.keys())
                dw = csv.DictWriter(sys.stdout, fieldnames=fieldnames)
                dw.writeheader()
            dw.writerow({k: v for k, v in row.items() if k in fieldnames})
            continue

        print(f"\n## Match {i} COSINE SIMILARITY: {sim}")
        dump(obj1, output_format)
        dump(obj2, output_format)


@main.command()
@path_option
@collection_option
@database_type_option
@click.argument("id")
def matches(id, path, collection, database_type):
    """Find matches for an ID.

    curategpt matches "Continuant" -p duckdb/objects.duckdb -c objects_a -D duckdb

    """
    db = get_store(database_type, path)
    # TODO: persist this in the database
    db.text_lookup = "label"
    obj = db.lookup(id, collection=collection)
    print(obj)
    results = db.matches(obj, collection=collection)
    i = 0
    for obj, distance, _meta in results:
        i += 1
        print(f"## ID:- {obj['id']}")
        print(f"## DISTANCE- {distance}")
        # print(yaml.dump(obj, sort_keys=False))


@main.command()
@path_option
@collection_option
@model_option
@limit_option
@database_type_option
@click.option(
    "--identifier-field",
    "-I",
    help="Field to use as identifier (defaults to id).",
)
@click.option(
    "--label-field",
    "-L",
    help="Field to use as label (defaults to label).",
)
@click.option("-l", "--limit", default=50, show_default=True, help="Number of candidate terms.")
@click.option(
    "--input-file",
    "-i",
    type=click.File("r"),
    help="Input file (one text per line).",
)
@click.option(
    "--split-sentences/--no-split-sentences",
    "-s/-S",
    default=False,
    show_default=True,
    help="Whether to split sentences.",
)
# choose from options in AnnotationMethod
@click.option(
    "--method",
    "-M",
    default=AnnotationMethod.INLINE.value,
    show_default=True,
    type=click.Choice([m for m in AnnotationMethod]),
    help="Annotation method.",
)
@click.option(
    "--prefix",
    multiple=True,
    help="Prefix(es) for candidate IDs.",
)
@click.option(
    "--category",
    multiple=True,
    help="Category/ies for candidate IDs.",
)
@click.argument("texts", nargs=-1)
def annotate(
    texts,
    path,
    model,
    collection,
    input_file,
    split_sentences,
    category,
    prefix,
    identifier_field,
    label_field,
    database_type,
    **kwargs,
):
    """Concept recognition."""
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if input_file:
        texts = [line.strip() for line in input_file]
    if model:
        extractor.model_name = model
    # TODO: persist this in the database
    cr = ConceptRecognitionAgent(knowledge_source=db, extractor=extractor)
    if prefix:
        cr.prefixes = list(prefix)
    categories = list(category) if category else None
    if identifier_field:
        cr.identifier_field = identifier_field
    if label_field:
        cr.label_field = label_field
    if split_sentences:
        new_texts = []
        for text in texts:
            for sentence in text.split("."):
                new_texts.append(sentence.strip())
        texts = new_texts
    for text in texts:
        ao = cr.annotate(text, collection=collection, categories=categories, **kwargs)
        dump(ao)
        print("---\n")


@main.command()
@path_option
@collection_option
@database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "--fields-to-predict",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@schema_option
@click.option(
    "--input",
    "-i",
    default=None,
    help="Input file to extract.",
)
@output_format_option
@click.argument("text", nargs=-1)
def extract(
    text,
    input,
    path,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    schema,
    output_format,
    database_type,
    **kwargs,
):
    """Extract a structured object from text.

    This uses RAG to provide the most relevant example objects
    from the collection to guide the extraction.

    Example:

        curategpt extract -c ont_foodon \
           "Chip butties are scottish delicacies consisting of \
            a buttered roll filled with deep fried potato wedges"

    """
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    agent = DatabaseAugmentedStructuredExtraction(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        agent.document_adapter = ChromaDBAdapter(docstore_path)
        agent.document_adapter_collection = docstore_collection
    if not text:
        if not input:
            raise ValueError("Must provide either text or input file.")
        text = list(open(input).readlines())
    text = "\n".join(text)
    ao = agent.extract(text, rules=rule, **filtered_kwargs)
    dump(ao.object, format=output_format)


@main.command()
@path_option
@collection_option
@database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "--fields-to-predict",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@click.option(
    "--output-directory",
    "-o",
    required=True,
)
@schema_option
@click.option(
    "--pubmed-id-file",
)
@click.argument("ids", nargs=-1)
def extract_from_pubmed(
    ids,
    pubmed_id_file,
    output_directory,
    path,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    schema,
    database_type,
    **kwargs,
):
    """Extract structured knowledge from a publication using its PubMed ID.

    For best results, use the PMID: prefix with PubMed IDs and the PMC: prefix with PMC IDs.
    Do not include the PMC prefix in the ID.

    Example:
    curategpt extract-from-pubmed -c ont_hp -o temp/ PMID:31851653

    See the `extract` command
    """
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    agent = DatabaseAugmentedStructuredExtraction(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        agent.document_adapter = ChromaDBAdapter(docstore_path)
        agent.document_adapter_collection = docstore_collection
    if not ids:
        if not pubmed_id_file:
            raise ValueError("Must provide either text or input file.")
        ids = [x.strip() for x in open(pubmed_id_file).readlines()]
    pmw = PubmedWrapper()
    output_directory = Path(output_directory)
    output_directory.mkdir(exist_ok=True, parents=True)
    for pmid in ids:
        pmid_esc = pmid.replace(":", "_")
        text = pmw.fetch_full_text(pmid)
        if not text:
            logging.warning(f"Could not fetch text for {pmid}")
            continue
        else:
            ao = agent.extract(text, rules=rule, **filtered_kwargs)
            with open(output_directory / f"{pmid_esc}.yaml", "w") as f:
                f.write(yaml.dump(ao.object, sort_keys=False))
            with open(output_directory / f"{pmid_esc}.txt", "w") as f:
                f.write(text)


@main.group()
def bootstrap():
    """Bootstrap schema or data.

    Starting with a general description or a LinkML schema,
    generate an initial version of a knowledge base.

    The config should be a yaml file with the following fields:
    kb_name: str
    description: str
    attributes: str
    main_class: str

    For example, this is a valid config:
    kb_name: lumber_kb
    description: A knowledge base for lumber
    attributes: source_tree
    main_class: Lumber_Type

    Examples:

    curategpt bootstrap schema -C config.yaml
    (This will generate a LinkML schema, based on the provided config.)

    curategpt bootstrap data -s schema.yaml
    (This will generate data based on the provided schema.
    The output of the previous command can be used as input for this command.)
    """


@bootstrap.command(name="schema")
@model_option
@click.option(
    "--config",
    "-C",
    required=True,
    help="path to yaml config",
)
def bootstrap_schema(config, model):
    """Bootstrap a knowledge base with LinkML schema."""
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    bootstrap_agent = BootstrapAgent(extractor=extractor)
    config_dict = yaml.safe_load(open(config))
    config = KnowledgeBaseSpecification(**config_dict)
    ao = bootstrap_agent.bootstrap_schema(config)
    dump(ao.object)


@bootstrap.command(name="data")
@model_option
@click.option(
    "--config",
    "-C",
    help="path to yaml config",
)
@click.option(
    "--schema",
    "-s",
    help="path to yaml linkml schema",
)
def bootstrap_data(config, schema, model):
    """Bootstrap a knowledge base with initial data."""
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    bootstrap_agent = BootstrapAgent(extractor=extractor)
    if config:
        config_dict = yaml.safe_load(open(config))
        config = KnowledgeBaseSpecification(**config_dict)
    else:
        config = None
    if schema:
        schema_dict = yaml.safe_load(open(schema))
    else:
        schema_dict = None
    yaml_str = bootstrap_agent.bootstrap_data(specification=config, schema=schema_dict)
    print(yaml_str)


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "-P", "--query-property", default="label", show_default=True, help="Property to use for query."
)
@click.option(
    "--fields-to-predict",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge. These are included in the prompt.",
)
@extract_format_option
@schema_option
@output_format_option
@click.argument("query")
def complete(
    query,
    path,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    query_property,
    schema,
    extract_format,
    output_format,
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Generate an entry from a query using object completion.

    Example:
    -------

        curategpt complete  -c obo_go "umbelliferose biosynthetic process"

    If the string looks like yaml (if it has a ':') then it will be parsed as yaml.

    E.g

        curategpt complete  -c obo_go "label: umbelliferose biosynthetic process"
        curategpt complete -p duckdb/objects.duckdb -c objects -D duckdb "label: more continuant"

    Pass ``--extract-format`` to make the extractor use a different internal representation
    when communicating to the LLM
    """
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if extract_format:
        extractor.serialization_format = extract_format
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dac = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dac.document_adapter = get_store(docstore_database_type, docstore_path)
        dac.document_adapter_collection = docstore_collection
    if ":" in query:
        query = yaml.safe_load(query)
    ao = dac.complete(query, context_property=query_property, rules=rule, **filtered_kwargs)
    dump(ao.object, format=output_format)


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "-P", "--query-property", default="label", show_default=True, help="Property to use for query."
)
@click.option(
    "--fields-to-predict",
    "-Z",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge. These are included in the prompt.",
)
@extract_format_option
@schema_option
@output_format_option
@click.option("--primary-key", help="Primary key for patch output.")
@click.argument("where", nargs=-1)
def update(
    where,
    path,
    collection,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    query_property,
    schema,
    extract_format,
    output_format,
    primary_key,
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Update an entry from a database using object completion.
    Example:
    -------
        curategpt update -X yaml --model gpt-4o -p db -c disease -Z description name: Asthma --primary-key name -t patch > patch.yaml
        curategpt update -X yaml --model gpt-4o -p duckdb/objects.duckdb -c objects -D duckdb -Z definition id: Continuant --primary-key id -t patch > patch.yaml

    """
    db = get_store(database_type, path)
    where_str = " ".join(where)
    where_q = yaml.safe_load(where_str)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None
    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if extract_format:
        extractor.serialization_format = extract_format
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dac = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dac.document_adapter = get_store(docstore_database_type, docstore_path)
        dac.document_adapter_collection = docstore_collection
    for obj, _s, _meta in db.find(where_q, collection=collection):
        # click.echo(f"{obj}")
        logging.debug(f"Updating {obj}")
        ao = dac.complete(
            obj,
            context_property=query_property,
            rules=rule,
            collection=collection,
            **filtered_kwargs,
        )
        if output_format == "yaml":
            click.echo("---")
        dump(ao.object, format=output_format, old_object=obj, primary_key=primary_key)


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "-P", "--query-property", default="label", show_default=True, help="Property to use for query."
)
@click.option(
    "--fields-to-predict",
    "-Z",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge. These are included in the prompt.",
)
@extract_format_option
@schema_option
@output_format_option
@click.option("--primary-key", help="Primary key for patch output.")
@click.argument("where", nargs=-1)
def review(
    where,
    path,
    collection,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    query_property,
    schema,
    extract_format,
    output_format,
    primary_key,
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Review entries.

    Example:
    -------

        curategpt review  -c obo_obi "{}" -Z definition -t patch \
          --primary-key original_id --rule "make definitions simple and easy to read by domain scientists. \
             At the same time, conform to genus-differentia style and OBO best practice."

    """
    where_str = " ".join(where)
    where_q = yaml.safe_load(where_str)
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if extract_format:
        extractor.serialization_format = extract_format
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dac = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dac.document_adapter = get_store(docstore_database_type, docstore_path)
        dac.document_adapter_collection = docstore_collection
    for obj, _s, _meta in db.find(where_q, collection=collection):
        logging.debug(f"Updating {obj}")
        ao = dac.review(
            obj,
            rules=rule,
            collection=collection,
            context_property=query_property,
            primary_key=primary_key,
            **filtered_kwargs,
        )
        if output_format == "yaml":
            print("---")
        dump(ao.object, format=output_format, old_object=obj, primary_key=primary_key)


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option(
    "-P", "--query-property", default="label", show_default=True, help="Property to use for query."
)
@click.option(
    "--fields-to-predict",
    multiple=True,
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@schema_option
@extract_format_option
@output_format_option
@click.argument("input_file")
def complete_multiple(
    input_file,
    path,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    query_property,
    schema,
    output_format,
    extract_format,
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Generate an entry from a query using object completion for multiple objects.

    Example:
    -------
        curategpt complete-multiple -c obo_go -P label terms.txt
    """
    # TODO: NOT TESTED
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor(serialization_format=extract_format)
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dac = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dac.document_adapter = get_store(docstore_database_type, docstore_path)
        dac.document_adapter_collection = docstore_collection
    with open(input_file) as f:
        queries = [l.strip() for l in f.readlines()]
        for query in queries:
            query = query.split("\t")[0]
            if ":" in query:
                query = yaml.safe_load(query)
            ao = dac.complete(query, context_property=query_property, rules=rule, **filtered_kwargs)
            print("---")
            dump(ao.object, format=output_format)


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@model_option
@limit_option
@click.option(
    "-P", "--query-property", default="label", show_default=True, help="Property to use for query."
)
@click.option(
    "--fields-to-predict",
    multiple=True,
)
@click.option(
    "--number-of-entries",
    "-N",
    default=5,
    show_default=True,
    help="number of entries to generate",
)
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@generate_background_option
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@schema_option
@extract_format_option
@output_format_option
def complete_auto(
    path,
    collection,
    docstore_path,
    docstore_collection,
    rule: List[str],
    model,
    query_property,
    schema,
    output_format,
    extract_format,
    number_of_entries,
    **kwargs,
):
    """
    Generate new KB entries, using model to choose new entities.

    Example:
    -------
        curategpt complete-auto -c obo_go -N 5
    """
    db = ChromaDBAdapter(path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor(serialization_format=extract_format)
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dac = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dac.document_adapter = ChromaDBAdapter(docstore_path)
        dac.document_adapter_collection = docstore_collection
    i = 0
    while i < number_of_entries:
        queries = dac.generate_queries(
            context_property=query_property, n=number_of_entries, collection=collection
        )
        logging.info(f"SUGGESTIONS: {queries}")
        if not queries:
            raise ValueError("No results")
        for query in queries:
            logging.info(f"SUGGESTION: {query}")
            ao = dac.complete(
                query,
                context_property=query_property,
                rules=rule,
                collection=collection,
                **filtered_kwargs,
            )
            print("---")
            dump(ao.object, format=output_format)
            i += 1


@main.command()
@path_option
@collection_option
@click.option(
    "-C/--no-C",
    "--conversation/--no-conversation",
    default=False,
    show_default=True,
    help="Whether to run in conversation mode.",
)
@model_option
@limit_option
@click.option("--field-to-predict", "-F", help="Field to predict")
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@click.option(
    "--generate-background/--no-generate-background",
    default=False,
    show_default=True,
    help="Whether to generate background knowledge.",
)
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@click.option(
    "--id-file",
    "-i",
    type=click.File("r"),
    help="File to read ids from.",
)
@click.option(
    "--missing-only/--no-missing-only",
    default=True,
    show_default=True,
    help="Only generate missing values.",
)
@schema_option
def complete_all(
    path,
    collection,
    docstore_path,
    docstore_collection,
    conversation,
    rule: List[str],
    model,
    field_to_predict,
    schema,
    id_file,
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Generate missing values for all objects

    Example:
    -------
        curategpt generate  -c obo_go
    """
    # TODO: NOT TESTED
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # TODO: generalize
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    dae = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        dae.document_adapter = get_store(docstore_database_type, docstore_path)
        dae.document_adapter_collection = docstore_collection
    object_ids = None
    if id_file:
        object_ids = [line.strip() for line in id_file.readlines()]
    it = dae.generate_all(
        collection=collection,
        field_to_predict=field_to_predict,
        rules=rule,
        object_ids=object_ids,
        **filtered_kwargs,
    )
    for pred in it:
        print(yaml.dump(pred.dict(), sort_keys=False))


@main.command()
@path_option
@collection_option
@database_type_option
@docstore_database_type_option
@click.option("--test-collection", "-T", required=True, help="Collection to use as the test set")
@click.option(
    "--hold-back-fields",
    "-F",
    required=True,
    help="Comma separated list of fields to predict in the test.",
)
@click.option(
    "--mask-fields",
    "-M",
    help="Comma separated list of fields to mask in the test.",
)
@model_option
@limit_option
@click.option(
    "--docstore-path",
    default=None,
    help="Path to a docstore to for additional unstructured knowledge.",
)
@click.option("--docstore-collection", default=None, help="Collection to use in the docstore.")
@click.option(
    "--generate-background/--no-generate-background",
    default=False,
    show_default=True,
    help="Whether to generate background knowledge.",
)
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@click.option(
    "--report-file",
    "-o",
    type=click.File("w"),
    help="File to write report to.",
)
@click.option(
    "--num-tests",
    default=None,
    show_default=True,
    help="Number (max) of tests to run.",
)
@schema_option
def generate_evaluate(
    path,
    docstore_path,
    docstore_collection,
    model,
    schema,
    test_collection,
    num_tests,
    hold_back_fields,
    mask_fields,
    rule: List[str],
    database_type,
    docstore_database_type,
    **kwargs,
):
    """
    Evaluate generate using a test set.

    Example:
    -------
        curategpt -v generate-evaluate -c cdr_training -T cdr_test -F statements -m gpt-4
    """
    db = get_store(database_type, path)
    if schema:
        schema_manager = SchemaProxy(schema)
    else:
        schema_manager = None

    # filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if schema_manager:
        db.schema_proxy = schema
        extractor.schema_proxy = schema_manager
    rage = DragonAgent(knowledge_source=db, extractor=extractor)
    if docstore_path or docstore_collection:
        rage.document_adapter = get_store(docstore_database_type, docstore_path)
        rage.document_adapter_collection = docstore_collection
    hold_back_fields = hold_back_fields.split(",")
    mask_fields = mask_fields.split(",") if mask_fields else []
    evaluator = DatabaseAugmentedCompletionEvaluator(
        agent=rage, fields_to_predict=hold_back_fields, fields_to_mask=mask_fields
    )
    results = evaluator.evaluate(test_collection, num_tests=num_tests, **kwargs)
    print(yaml.dump(results.dict(), sort_keys=False))


@main.command()
@path_option
@collection_option
@click.option(
    "--hold-back-fields",
    "-F",
    help="Comma separated list of fields to predict in the test.",
)
@click.option(
    "--mask-fields",
    "-M",
    help="Comma separated list of fields to mask in the test.",
)
@model_option
@limit_option
@click.option(
    "--generate-background/--no-generate-background",
    default=False,
    show_default=True,
    help="Whether to generate background knowledge.",
)
@click.option(
    "--rule",
    multiple=True,
    help="Rule to use for generating background knowledge.",
)
@click.option(
    "--report-file",
    "-o",
    type=click.File("w"),
    help="File to write report to.",
)
@click.option(
    "--num-testing",
    default=None,
    show_default=True,
    help="Number (max) of tests to run.",
)
@click.option(
    "--working-directory",
    "-W",
    help="Working directory to use.",
)
@click.option(
    "--fresh/--no-fresh",
    default=False,
    show_default=True,
    help="Whether to rebuild test/train collections.",
)
@generate_background_option
@click.argument("tasks", nargs=-1)
def evaluate(
    tasks,
    working_directory,
    path,
    model,
    generate_background,
    num_testing,
    hold_back_fields,
    mask_fields,
    rule: List[str],
    collection,
    **kwargs,
):
    """
    Evaluate given a task configuration.

    Example:
    -------
        curategpt evaluate src/curategpt/conf/tasks/bio-ont.tasks.yaml
    """
    normalized_tasks = []
    for task in tasks:
        if ":" in task:
            task = yaml.safe_load(task)
        else:
            task = yaml.safe_load(open(task))
        if isinstance(task, list):
            normalized_tasks.extend(task)
        else:
            normalized_tasks.append(task)
    for task in normalized_tasks:
        task_obj = Task(**task)
        if path:
            task_obj.path = path
        if working_directory:
            task_obj.working_directory = working_directory
        if collection:
            task_obj.source_collection = collection
        if model:
            task_obj.model_name = model
        if hold_back_fields:
            task_obj.hold_back_fields = hold_back_fields.split(",")
        if mask_fields:
            task_obj.mask_fields = mask_fields.split(",")
        if num_testing is not None:
            task_obj.num_testing = int(num_testing)
        if generate_background:
            task_obj.generate_background = generate_background
        if rule:
            # TODO
            task_obj.rules = rule
        result = run_task(task_obj, **kwargs)
        print(yaml.dump(result.dict(), sort_keys=False))


@main.command()
@database_type_option
@path_option
@click.option("--collections", required=True)
@click.option("--models", default="gpt-3.5-turbo")
@click.option("--fields-to-mask", default="id,original_id")
@click.option("--fields-to-predict", required=True)
@click.option("--num-testing", default=50, show_default=True)
@click.option("--background", default="false", show_default=True)
def evaluation_config(collections, models, fields_to_mask, fields_to_predict, background, **kwargs):
    tasks = []
    # TODO: is there anything to do?
    # db = get_store(kwargs["database_type"], kwargs["path"])
    for collection in collections.split(","):
        for model in models.split(","):
            for fp in fields_to_predict.split(","):
                for bg in background.split(","):
                    tc = Task(
                        source_db_path="db",
                        target_db_path="db",
                        model_name=model,
                        source_collection=collection,
                        fields_to_predict=[fp],
                        fields_to_mask=fields_to_mask.split(","),
                        generate_background=json.loads(bg),
                        stratified_collection=StratifiedCollection(
                            training_set_collection=f"{collection}_training",
                            testing_set_collection=f"{collection}_testing",
                        ),
                        **kwargs,
                    )
                    tasks.append(tc.dict(exclude_unset=True))
    print(yaml.dump(tasks, sort_keys=False))


@main.command()
@click.option(
    "--include-expected/--no-include-expected",
    "-E",
    default=False,
    show_default=True,
)
@click.argument("files", nargs=-1)
def evaluation_compare(files, include_expected=False):
    """Compare evaluation results."""
    dfs = []
    predicted_cols = []
    other_cols = []
    differentia_col = "method"
    for f in files:
        df = pd.read_csv(f, sep="\t", comment="#")
        df[differentia_col] = f
        if include_expected:
            include_expected = False
            base_df = df.copy()
            base_df[differentia_col] = "source"
            for c in base_df.columns:
                if c.startswith("expected_"):
                    new_col = c.replace("expected_", "predicted_")
                    base_df[new_col] = base_df[c]
            dfs.append(base_df)
        dfs.append(df)
        for c in df.columns:
            if c in predicted_cols or c in other_cols:
                continue
            if c.startswith("predicted_"):
                predicted_cols.append(c)
            else:
                other_cols.append(c)
    df = pd.concat(dfs)
    # df = pd.melt(df, id_vars=["masked_id", "file"], value_vars=["predicted_definition"])
    df = pd.melt(df, id_vars=list(other_cols), value_vars=list(predicted_cols))
    df = df.sort_values(by=list(other_cols))
    df.to_csv(sys.stdout, sep="\t", index=False)


@main.command()
@click.option(
    "--system",
    "-s",
    help="System gpt prompt to use.",
)
@click.option(
    "--prompt",
    "-p",
    default="What is the definition of {column}?",
)
@model_option
@click.argument("file")
def multiprompt(file, model, system, prompt):
    if model is None:
        model = "gpt-3.5-turbo"
    model_obj = get_model(model)
    with open(file) as f:
        for row in csv.DictReader(f, delimiter="\t"):
            resp = model_obj.prompt(system=system, prompt=prompt.format(**row)).text()
            resp = resp.replace("\n", " ")
            print("\t".join(list(row.values()) + [resp]))


@main.command()
@collection_option
@path_option
@model_option
@database_type_option
@click.option(
    "--show-references/--no-show-references",
    default=True,
    show_default=True,
    help="Whether to show references.",
)
@click.option(
    "_continue",
    "-C",
    "--continue",
    is_flag=True,
    flag_value=-1,
    help="Continue the most recent conversation.",
)
@click.option(
    "conversation_id",
    "--cid",
    "--conversation",
    help="Continue the conversation with the given ID.",
)
@click.argument("query")
def ask(query, path, collection, model, show_references, _continue, conversation_id, database_type):
    """Chat with data in a collection.

    Example:

        curategpt ask -c obo_go "What are the parts of the nucleus?"

    """
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    conversation = None
    if conversation_id or _continue:
        # Load the conversation - loads most recent if no ID provided
        try:
            conversation = load_conversation(conversation_id)
            print(f"CONTINUING CONVERSATION {conversation}")
        except UnknownModelError as ex:
            raise click.ClickException(str(ex)) from ex
    chatbot = ChatAgent(path)
    chatbot.extractor = extractor
    chatbot.knowledge_source = db
    response = chatbot.chat(query, collection=collection, conversation=conversation)
    show_chat_response(response, show_references)


@main.command()
@click.option("--patch", help="Patch file.")
@click.option("--primary-key", help="Primary key for patch output")
@click.argument("input_file")
def apply_patch(input_file, patch, primary_key):
    """Apply a patch to a file/KB.

    This can be executed after `update`
    """
    objs = list(yaml.safe_load_all(open(input_file)))
    logging.info(f"Applying patch to {len(objs)} objects")
    patch = yaml.safe_load(open(patch))
    if primary_key:
        for inner_obj in objs:
            pk_val = inner_obj[primary_key]
            if pk_val in patch:
                actual_patch = patch[pk_val]
                logging.debug(f"Applying: {actual_patch}")
                jsonpatch.apply_patch(inner_obj, actual_patch, in_place=True)
    else:
        for obj in objs:
            jsonpatch.apply_patch(obj, patch, in_place=True)
    logging.info(f"Writing patch output for {len(objs)} objects")
    for obj in objs:
        print("---")
        print(yaml.dump(obj, sort_keys=False))


@main.command()
@collection_option
@path_option
@model_option
@database_type_option
@click.option(
    "--show-references/--no-show-references",
    default=True,
    show_default=True,
    help="Whether to show references.",
)
@click.option(
    "_continue",
    "-C",
    "--continue",
    is_flag=True,
    flag_value=-1,
    help="Continue the most recent conversation.",
)
@click.option(
    "conversation_id",
    "--cid",
    "--conversation",
    help="Continue the conversation with the given ID.",
)
@click.option(
    "--select",
    help="jsonpath expression to select objects from the input file.",
)
@click.argument("query")
def citeseek(
    query,
    path,
    collection,
    model,
    show_references,
    _continue,
    select,
    conversation_id,
    database_type,
):
    """Find citations for an object or statement.

    You can pass in a statement directly as an argument

    Example:

        curategpt -citeseek --model gpt-4o "Ivermectin cures COVID-19"

    Returns:

    .. code-block:: yaml

      - reference: PMID:35657909
        supports: REFUTE
        snippet: 'COVID-19 update: NIH recommends against ivermectin.'
        explanation: This reference clearly states that the National Institutes of Health
          (NIH) recommends against the use of ivermectin for COVID-19 treatment.
      - reference: PMID:35225114
        supports: REFUTE
        snippet: Due to concern for adverse events, specifically neurotoxicity, as well
          as a paucity of supporting evidence, the use of ivermectin as a routine treatment
          or preventive measure for COVID-19 infection is not recommended at this time.
        explanation: The reference indicates concerns about adverse events and a lack
          of supporting evidence, leading to the conclusion that ivermectin is not recommended
          for COVID-19 treatment.

    You can also pass a YAML file as an argument. All assertions within the YAML file
    will be individually checked.

    Example:

        curategpt -v  citeseek --model gpt-4o tests/input/citeseek-test.yaml
        curategpt -v citeseek --model gpt-4o -p duckdb/objects.duckdb -D duckdb "Continuant"


    """
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if not collection or collection == "pubmed":
        chatbot = PubmedWrapper(local_store=db, extractor=extractor)
    else:
        chatbot = ChatAgent(db, extractor=extractor, knowledge_source_collection=collection)
    ea = EvidenceAgent(chat_agent=chatbot)
    if Path(query).exists():
        try:
            logging.info(f"Testing if query is a file: {query}")
            parsed_obj = list(yaml.safe_load_all(open(query)))
            if isinstance(parsed_obj, list):
                objs = parsed_obj
            else:
                objs = [parsed_obj]
            logging.info(f"Loaded {len(objs)} objects from {query}")
            if select:
                logging.info(f"Selecting objects using {select}")
                # TODO: DRY
                import jsonpath_ng as jp

                path_expr = jp.parse(select)
                new_objs = []
                for obj in objs:
                    for match in path_expr.find(obj):
                        logging.debug(f"Match: {match.value}")
                        if isinstance(match.value, list):
                            new_objs.extend(match.value)
                        else:
                            new_objs.append(match.value)
                objs = new_objs
                logging.info(f"New {len(objs)} objects from {select}")
            for obj in objs:
                enhanced_obj = ea.find_evidence_complex(obj)
                print("---")
                print(yaml.dump(enhanced_obj, sort_keys=False), flush=True)
            return
        except Exception as ex:
            print(f"Error reading {query}: {ex}")
    logging.info(f"Query: {query}")
    response = ea.find_evidence_simple(query)
    print(yaml.dump(response, sort_keys=False))


@main.command()
@collection_option
@path_option
@model_option
@database_type_option
@click.option("--view", "-V", help="Name of the wrapper to use.")
@click.option("--name-field", help="Field for names.")
@click.option("--description-field", help="Field for descriptions.")
@click.option("--system-prompt", help="System gpt prompt to use.")
@click.argument("ids", nargs=-1)
def summarize(ids, path, collection, model, view, database_type, **kwargs):
    """
    Summarize a list of objects.

    Retrieves objects by ID from a knowledge source or wrapper and summarizes them.

    (this is a partial implementation of TALISMAN using CurateGPT)

    Example:
    -------
      curategpt summarize --model llama-2-7b-chat -V alliance_gene \
        --name-field symbol --description-field automatedGeneSynopsis \
        --system-prompt "What functions do these genes share?" \
        HGNC:3239 HGNC:7632 HGNC:4458 HGNC:9439 HGNC:29427 \
        HGNC:1160  HGNC:26270 HGNC:24682 HGNC:7225 HGNC:13797 \
        HGNC:9118  HGNC:6396  HGNC:9179 HGNC:25358
    """
    # TODO: with llama-2-7b-chat it's not working (ChunkedEncodingError)
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    if view:
        db = get_wrapper(view)
    agent = SummarizationAgent(db, extractor=extractor, knowledge_source_collection=collection)
    response = agent.summarize(ids, **kwargs)
    print("# Response:")
    click.echo(response)


@main.command()
def plugins():
    "List installed plugins"
    print(yaml.dump(get_plugins()))


@main.group()
def collections():
    "Operate on collections in the store."


@collections.command(name="list")
@click.option(
    "--minimal/--no-minimal",
    default=False,
    show_default=True,
    help="Whether to show minimal information.",
)
@click.option(
    "--derived/--no-derived",
    default=True,
    show_default=True,
    help="Whether to show derived information.",
)
@click.option(
    "--peek/--no-peek",
    default=False,
    show_default=True,
    help="Whether to peek at the first few entries of the collection.",
)
@database_type_option
@path_option
def list_collections(database_type, path, peek: bool, minimal: bool, derived: bool):
    """List all collections"""
    logging.info(f"Listing collections in {path}")
    db = get_store(database_type, path)
    logging.info(f"Initialized: {db}")
    for cn in db.list_collection_names():
        if minimal:
            print(f"## Collection: {cn}")
            continue
        cm = db.collection_metadata(cn, include_derived=derived)
        if database_type == "chromadb":
            # TODO: make get_or_create abstract and implement in DBAdapter?
            c = db.client.get_or_create_collection(cn)
            print(f"## Collection: {cn} N={c.count()} meta={c.metadata} // {cm}")
            if peek:
                r = c.peek()
                for id_ in r["ids"]:
                    print(f" - {id_}")
        if database_type == "duckdb":
            print(f"## Collection: {cm}")
            if peek:
                # making sure if o[id] finds nothing we get the full obj
                r = list(db.peek(cn))
                for o, _, _ in r:
                    if "id" in o:
                        print(f" - {o['id']}")
                    else:
                        print(f" - {o}")


@collections.command(name="delete")
@collection_option
@path_option
@database_type_option
def delete_collection(path, collection, database_type):
    """Delete a collections."""
    db = get_store(database_type, path)
    db.remove_collection(collection)


@collections.command(name="peek")
@collection_option
@database_type_option
@limit_option
@path_option
def peek_collection(path, collection, database_type, **kwargs):
    """Inspect a collection."""
    logging.info(f"Peeking at {collection} in {path}")
    db = get_store(database_type, path)
    if database_type == "chromadb":
        for obj in db.peek(collection, **kwargs):
            print(yaml.dump(obj, sort_keys=False))
    if database_type == "duckdb":
        for obj in db.peek(collection, **kwargs):
            print(yaml.dump(obj.metadatas, sort_keys=False))


@collections.command(name="dump")
@collection_option
@click.option("-o", "--output", default="-")
@click.option("--metadata-to-file", type=click.File("w"), default=None)
@click.option("--format", "-t", default="json", show_default=True)
@click.option("--include", "-I", multiple=True, help="Include a field.")
@path_option
@database_type_option
def dump_collection(path, collection, output, database_type, **kwargs):
    """
    Dump a collection to disk.

    There are two flavors of format:

    - streaming, flat lists of objects (e.g. jsonl)
    - complete (e.g json)

    with streaming formats it's necessary to also provide `--metadata-to-file` since
    the metadata header won't fit into the line-based formats.

    Example:

        curategpt collections dump -p stagedb -D chromadb -c objects -o objects.cur.json

    Example:

        curategpt collections dump  -c ont_cl -o cl.cur.jsonl -t jsonl --metadata-to-file cl.meta.json

    """
    logging.info(f"Dumping {collection} in {path}")
    db = get_store(database_type, path)
    db.dump(collection, to_file=output, **kwargs)


@collections.command(name="copy")
@collection_option
@click.option("--target-path")
@path_option
@database_type_option
def copy_collection(path, collection, target_path, database_type, **kwargs):
    """
    Copy a collection from one path to another.

    Example:

        curategpt collections copy -p stagedb --target-path db -c my_collection
    """
    # TODO: not tested
    logging.info(f"Copying {collection} in {path} to {target_path}")
    db = get_store(database_type, path)
    target = get_store(database_type, target_path)
    db.dump_then_load(collection, target=target)


@collections.command(name="split")
@collection_option
@database_type_option
@click.option(
    "--derived-collection-base",
    help=(
        "Base name for derived collections. Will be suffixed with _train, _test, _val."
        "If not provided, will use the same name as the original collection."
    ),
)
@model_option
@click.option(
    "--num-training",
    type=click.INT,
    help="Number of training examples to keep.",
)
@click.option(
    "--num-testing",
    type=click.INT,
    help="Number of testing examples to keep.",
)
@click.option(
    "--num-validation",
    default=0,
    show_default=True,
    type=click.INT,
    help="Number of validation examples to keep.",
)
@click.option(
    "--test-id-file",
    type=click.File("r"),
    help="File containing IDs of test objects.",
)
@click.option(
    "--ratio",
    type=click.FLOAT,
    help="Ratio of training to testing examples.",
)
@click.option(
    "--fields-to-predict",
    "-F",
    required=True,
    help="Comma separated list of fields to predict in the test. Candidate objects must have these fields.",
)
@click.option(
    "--output-path",
    "-o",
    required=True,
    help="Path to write the new store.",
)
@path_option
def split_collection(
    path,
    collection,
    derived_collection_base,
    output_path,
    model,
    test_id_file,
    database_type,
    **kwargs,
):
    """
    Split a collection into test/train/validation.

    Example:
    -------
        curategpt -v collections split -c hp --num-training 10 --num-testing 20

    The above populates 2 new collections: hp_training and hp_testing.

    This can be run as a pre-processing step for generate-evaluate.
    """
    # TODO: not tested
    db = get_store(database_type, path)
    if test_id_file:
        kwargs["testing_identifiers"] = [line.strip().split()[0] for line in test_id_file]
        logging.info(
            f"Using {len(kwargs['testing_identifiers'])} testing identifiers from {test_id_file.name}"
        )
        logging.info(f"First 10: {kwargs['testing_identifiers'][:10]}")
    sc = stratify_collection(db, collection, **kwargs)
    output_db = get_store(database_type, output_path)
    if not derived_collection_base:
        derived_collection_base = collection
    for sn in ["training", "testing", "validation"]:
        cn = f"{derived_collection_base}_{sn}"
        output_db.remove_collection(cn, exists_ok=True)
        objs = getattr(sc, f"{sn}_set", [])
        logging.info(f"Writing {len(objs)} objects to {cn}")
        output_db.insert(objs, collection=cn, model=model)


@collections.command(name="set")
@collection_option
@path_option
@database_type_option
@click.argument("metadata_yaml")
def set_collection_metadata(path, collection, database_type, metadata_yaml):
    """Set metadata for a collection."""
    # TODO: not tested
    db = get_store(database_type, path)
    db.update_collection_metadata(collection, **yaml.safe_load(metadata_yaml))


@main.group()
def ontology():
    "Use the ontology model"


@ontology.command(name="index")
@path_option
@collection_option
@model_option
@append_option
@database_type_option
@click.option(
    "--branches",
    "-b",
    help="Comma separated list node IDs representing branches to index.",
)
@click.option(
    "--index-fields",
    help="Fields to index; comma separated",
)
@click.argument("ont")
def index_ontology_command(
    ont, path, collection, append, model, index_fields, branches, database_type, **kwargs
):
    """
    Index an ontology.

    Example:
    -------
        curategpt ontology index -c obo_hp $db/hp.db -D duckdb
        curategpt ontology index -p stagedb/duck.db -c ont-hp sqlite:obo:hp -D duckdb

    """

    s = time.time()
    oak_adapter = get_adapter(ont)
    view = OntologyWrapper(oak_adapter=oak_adapter)
    if branches:
        view.branches = branches.split(",")
    db = get_store(database_type, path)
    db.text_lookup = view.text_field
    if index_fields:
        fields = index_fields.split(",")

        # print(f"Indexing fields: {fields}")

        def _text_lookup(obj: Dict):
            vals = [str(obj.get(f)) for f in fields if f in obj]
            return " ".join(vals)

        db.text_lookup = _text_lookup
    if not append:
        db.remove_collection(collection, exists_ok=True)
    click.echo(f"Indexing {len(list(view.objects()))} objects")
    db.insert(view.objects(), collection=collection, model=model)
    db.update_collection_metadata(collection, object_type="OntologyClass")
    e = time.time()
    click.echo(f"Indexed {len(list(view.objects()))} in {e - s} seconds")


@main.group()
def embeddings():
    """Command group for handling embeddings."""
    pass


def download_file(url):
    """
    Helper function to download a file from a URL to a temporary file.
    """
    local_filename = tempfile.mktemp()
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename


def load_embeddings_from_file(file_path, embedding_format=None):
    """
    Helper function to load embeddings from a file. Supports Parquet and CSV formats.
    """
    if (
        file_path.endswith(".parquet")
        or file_path.endswith(".parquet.gz")
        or embedding_format == "parquet"
    ):
        df = pd.read_parquet(file_path)
    elif file_path.endswith(".csv") or file_path.endswith(".csv.gz") or embedding_format == "csv":
        df = pd.read_csv(file_path)
    else:
        raise ValueError("Unsupported file type. Only Parquet and CSV files are supported.")
    return df.to_dict(orient="records")


@embeddings.command(name="load")
@path_option
@collection_option
@model_option
@append_option
@database_type_option
@click.option(
    "--embedding-format",
    "-f",
    type=click.Choice(["parquet", "csv"]),
    help="Format of the input file",
)
@click.argument("file_or_url")
def load_embeddings(path, collection, append, embedding_format, model, file_or_url, database_type):
    """
    Index embeddings from a local file or URL into a ChromaDB collection.
    """
    # TODO: not tested (its not really the usecase of the adapters, at least you wanna have
    # ids or belonging objects? otherwise insert will calc that on the fly
    # consider using add (working for chroma at least) and add to objects ?
    # Check if file_or_url is a URL
    if file_or_url.startswith("http://") or file_or_url.startswith("https://"):
        print(f"Downloading file from URL: {file_or_url}")
        file_path = download_file(file_or_url)
    else:
        file_path = file_or_url

    print(f"Loading embeddings from file: {file_path}")
    embeddings = load_embeddings_from_file(file_path, embedding_format)

    # Initialize the database adapter
    db = get_store(database_type, path)
    if append:
        if collection in db.list_collection_names():
            print(f"Collection '{collection}' already exists. Adding to the existing collection.")
    else:
        db.remove_collection(collection, exists_ok=True)

    # Insert embeddings into the collection
    db.insert(embeddings, model=model, collection=collection)
    print(f"Successfully indexed embeddings into collection '{collection}'.")


@embeddings.command(name="upload")
@path_option
@collection_option
@click.option(
    "--repo-id",
    required=True,
    help="Repository ID on Hugging Face, e.g., 'biomedical-translator/[repo_name]'.",
)
@click.option("--private/--public", default=False, help="Whether the repository should be private.")
@click.option("--adapter", default="huggingface", help="Adapter to use for uploading embeddings.")
@database_type_option
def upload_embeddings(path, collection, repo_id, private, adapter, database_type):
    """
    Upload embeddings and their metadata from a specified collection to a repository,
    e.g. huggingface.

    Example:
        curategpt embeddings upload --repo-id biomedical-translator/my_repo --collection my_collection
    """
    db = get_store(database_type, path)

    try:
        objects = list(db.fetch_all_objects_memory_safe(collection=collection))
        metadata = db.collection_metadata(collection)
    except Exception as e:
        print(f"Error accessing collection '{collection}' from database: {e}")
        return

    if adapter == "huggingface":
        agent = HuggingFaceAgent()
    else:
        raise ValueError(
            f"Unsupported adapter: {adapter} " f"currently only huggingface adapter is supported"
        )
    try:
        agent.upload(objects=objects, metadata=metadata, repo_id=repo_id, private=private)
    except Exception as e:
        print(f"Error uploading collection to {repo_id}: {e}")


@main.group()
def view():
    "Virtual store/wrapper"


@view.command(name="objects")
@click.option("--view", "-V", required=True, help="Name of the wrapper to use.")
@click.option("--source-locator")
@click.option("--settings", help="YAML settings for the wrapper.")
@init_with_option
@click.argument("object_ids", nargs=-1)
def view_objects(view, init_with, settings, object_ids, **kwargs):
    """
    View objects in a virtual store.

    Example:
    -------
        curategpt view objects -V filesystem --init-with "root_directory: /path/to/data"

    """
    if init_with:
        for k, v in yaml.safe_load(init_with).items():
            kwargs[k] = v
    if settings:
        for k, v in yaml.safe_load(settings).items():
            kwargs[k] = v
    vstore = get_wrapper(view, **kwargs)
    if object_ids:
        for obj in vstore.objects(object_ids=object_ids):
            print(yaml.dump(obj, sort_keys=False))
    else:
        for obj in vstore.objects():
            print(yaml.dump(obj, sort_keys=False))
            print("---")


@view.command(name="unwrap")
@click.option("--view", "-V", required=True, help="Name of the wrapper to use.")
@click.option("--source-locator")
@path_option
@collection_option
@output_format_option
@database_type_option
@click.argument("input_file")
def unwrap_objects(input_file, view, path, collection, output_format, database_type, **kwargs):
    """
    Unwrap objects back to source schema.

    Example:
    -------

    Todo: duckdb adapter uses get_raw_objects, could impl same in chromadb adapter

    ----

    """
    vstore = get_wrapper(view, **kwargs)
    store = get_store(database_type, path)
    store.set_collection(collection)
    with open(input_file) as f:
        objs = yaml.safe_load_all(f)
        unwrapped = vstore.unwrap_object(objs, store=store)
        dump(unwrapped, output_format)


@view.command(name="search")
@click.option("--view", "-V")
@click.option("--source-locator")
@model_option
@limit_option
@init_with_option
@click.argument("query")
def view_search(query, view, model, init_with, limit, **kwargs):
    """Search in a virtual store."""
    if init_with:
        for k, v in yaml.safe_load(init_with).items():
            kwargs[k] = v
    vstore: BaseWrapper = get_wrapper(view, **kwargs)
    vstore.extractor = BasicExtractor(model_name=model)
    for obj, _dist, _ in vstore.search(query, limit=limit):
        print(yaml.dump(obj, sort_keys=False))


@view.command(name="index")
@path_option
@collection_option
@click.option("--view", "-V")
@click.option("--source-locator")
@batch_size_option
@model_option
@init_with_option
@append_option
@database_type_option
def view_index(
    view, path, append, collection, model, init_with, batch_size, database_type, **kwargs
):
    """Populate an index from a view.
    curategpt -v index -p stagedb --batch-size 10 -V hpoa  -c hpoa -m openai:  (that uses chroma by default)
    curategpt -v index -p stagedb/hpoa.duckdb --batch-size 10 -V hpoa  -c hpoa -m openai: -D duckdb

    """
    if init_with:
        for k, v in yaml.safe_load(init_with).items():
            kwargs[k] = v
    wrapper: BaseWrapper = get_wrapper(view, **kwargs)
    store = get_store(database_type, path)

    if not append:
        if collection in store.list_collection_names():
            store.remove_collection(collection)
    objs = wrapper.objects()
    store.insert(objs, model=model, collection=collection, batch_size=batch_size)


@view.command(name="ask")
@click.option("--view", "-V")
@click.option("--source-locator")
@limit_option
@model_option
@click.option(
    "--expand/--no-expand",
    default=True,
    show_default=True,
    help="Whether to expand the search term using an LLM.",
)
@click.argument("query")
def view_ask(query, view, model, limit, expand, **kwargs):
    """Ask a knowledge source wrapper."""
    vstore: BaseWrapper = get_wrapper(view)
    vstore.extractor = BasicExtractor(model_name=model)
    chatbot = ChatAgent(knowledge_source=vstore)
    response = chatbot.chat(query, limit=limit, expand=expand)
    show_chat_response(response, True)


@main.group()
def pubmed():
    "Use pubmed"


@pubmed.command(name="search")
@collection_option
@path_option
@model_option
@database_type_option
@click.option(
    "--expand/--no-expand",
    default=True,
    show_default=True,
    help="Whether to expand the search term using an LLM.",
)
@click.argument("query")
def pubmed_search(query, path, model, database_type, **kwargs):
    pubmed = PubmedWrapper()
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    pubmed.extractor = extractor
    pubmed.local_store = db
    results = pubmed.search(query, **kwargs)
    i = 0
    for obj, distance, _ in results:
        i += 1
        print(f"## {i} DISTANCE: {distance}")
        print(yaml.dump(obj, sort_keys=False))


@pubmed.command(name="ask")
@collection_option
@path_option
@model_option
@limit_option
@database_type_option
@click.option(
    "--show-references/--no-show-references",
    default=True,
    show_default=True,
    help="Whether to show references.",
)
@click.option(
    "--expand/--no-expand",
    default=True,
    show_default=True,
    help="Whether to expand the search term using an LLM.",
)
@click.argument("query")
def pubmed_ask(query, path, model, show_references, database_type, **kwargs):
    pubmed = PubmedWrapper()
    db = get_store(database_type, path)
    extractor = BasicExtractor()
    if model:
        extractor.model_name = model
    pubmed.extractor = extractor
    pubmed.local_store = db
    response = pubmed.chat(query, **kwargs)
    click.echo(response.formatted_body)
    if show_references:
        print("# References:")
        for ref, ref_text in response.references.items():
            print(f"## {ref}")
            print(ref_text)


if __name__ == "__main__":
    main()



"""EUtils-based wrapper for studies in NCBI."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, List

from curategpt.wrappers.literature.eutils_wrapper import EUtilsWrapper

logger = logging.getLogger(__name__)


@dataclass
class ClinVarWrapper(EUtilsWrapper):
    """
    A wrapper to provide a search facade over NCBI clinvar.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "clinvar"

    eutils_db: ClassVar[str] = "clinvar"

    id_prefix: ClassVar[str] = "ClinVar"

    default_object_type = "Variant"

    fetch_tool = "esummary"

    def objects_from_dict(self, results: Dict) -> List[Dict]:
        objs = []
        for r in results["eSummaryResult"]["DocumentSummarySet"]["DocumentSummary"]:
            obj = {}
            obj["id"] = "clinvar:" + r["accession"]
            obj["clinical_significance"] = r["germline_classification"]["description"]
            obj["clinical_significance_status"] = r["germline_classification"]["review_status"]
            obj["gene_sort"] = r["gene_sort"]
            if "genes" in r and r["genes"]:
                if "gene" in r["genes"]:
                    genes = r["genes"]["gene"]
                    if not isinstance(genes, list):
                        genes = [genes]
                    obj["genes"] = [self._gene_from_dict(g) for g in genes]
            obj["type"] = r["obj_type"]
            obj["protein_change"] = r["protein_change"]
            obj["title"] = r["title"]
            obj["traits"] = [
                self._trait_from_dict(t) for t in r.get("trait_set", {}).get("trait", []) if isinstance(t, dict)
            ]
            objs.append(obj)
        return objs

    def _gene_from_dict(self, gene: Dict) -> Dict:
        return {
            "id": "NCBIGene:" + gene["GeneID"],
            "symbol": gene["symbol"],
        }

    def _trait_from_dict(self, trait: Dict) -> Dict:
        obj = {
            "name": trait["trait_name"],
        }
        xrefsp = trait.get("trait_xrefs", None)
        if xrefsp:
            xrefs = xrefsp["trait_xref"]
            if not isinstance(xrefs, list):
                xrefs = [xrefs]
            obj["xrefs"] = [self._xref_from_dict(x) for x in xrefs]
        return obj

    def _xref_from_dict(self, xref: Dict) -> str:
        db_id = xref["db_id"]
        db_source = xref["db_source"]
        if db_id.lower().startswith(db_source.lower()):
            # no bananas
            return db_id
        else:
            return f"{db_source}:{db_id}"



"""
CurateGPT Extractors.

These handle connections to (remote or local) LLMs,
and can also extract structured objects from text.

* Base class: :class:`Extractor`
"""

from .basic_extractor import BasicExtractor
from .extractor import AnnotatedObject, Extractor
from .openai_extractor import OpenAIExtractor
from .recursive_extractor import RecursiveExtractor

__all__ = [
    "BasicExtractor",
    "AnnotatedObject",
    "Extractor",
    "RecursiveExtractor",
    "OpenAIExtractor",
]



from typing import List, Tuple

import numpy as np

from curategpt.utils.search import logger

LOL = List[List[float]]


def compute_cosine_similarity(list1: LOL, list2: LOL) -> np.ndarray:
    """
    Compute cosine similarity between two lists of vectors.

    Result is a two column vector sim[ROW][COL] where ROW is from list1 and COL is from list2.

    :param list1:
    :param list2:
    :return:
    """
    # Convert lists to numpy arrays
    matrix1 = np.array(list1)
    matrix2 = np.array(list2)

    # Normalize the vectors in both matrices
    matrix1_norm = matrix1 / np.linalg.norm(matrix1, axis=1)[:, np.newaxis]
    matrix2_norm = matrix2 / np.linalg.norm(matrix2, axis=1)[:, np.newaxis]

    # Compute dot products (resulting in cosine similarity values)
    cosine_similarity_matrix = np.dot(matrix1_norm, matrix2_norm.T)

    return cosine_similarity_matrix


def top_matches(cosine_similarity_matrix: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Find the top match for each row in the cosine similarity matrix.

    :param cosine_similarity_matrix:
    :return:
    """
    # Find the index of the maximum value in each row
    top_match_indices = np.argmax(cosine_similarity_matrix, axis=1)

    # Find the maximum similarity value in each row
    top_match_values = np.amax(cosine_similarity_matrix, axis=1)

    return top_match_indices, top_match_values


def top_n_matches(
    cosine_similarity_matrix: np.ndarray, n: int = 10
) -> Tuple[np.ndarray, np.ndarray]:
    # Find the indices that would sort each row in descending order
    sorted_indices = np.argsort(-cosine_similarity_matrix, axis=1)

    # Take the first n indices from the sorted indices to get the top n matches
    top_n_indices = sorted_indices[:, :n]

    # Take the first n values from the sorted values to get the top n match values
    top_n_values = -np.sort(-cosine_similarity_matrix, axis=1)[:, :n]

    return top_n_indices, top_n_values


def mmr_diversified_search(
    query_vector: np.ndarray, document_vectors: List[np.ndarray], relevance_factor=0.5, top_n=None
) -> List[int]:
    """
    Perform diversified search using Maximal Marginal Relevance (MMR).

    Parameters
    ----------
    - query_vector: The vector representing the query.
    - document_vectors: The vectors representing the documents.
    - lambda_: Balance parameter between relevance and diversity.
    - top_n: Number of results to return. If None, return all.

    Returns
    -------
    - List of indices representing the diversified order of documents.
    """
    # If no specific number of results is specified, return all
    if top_n is None:
        top_n = len(document_vectors)

    if top_n == 0:
        return []

    # Calculate cosine similarities between query and all documents
    norms_query = np.linalg.norm(query_vector)
    norms_docs = np.linalg.norm(document_vectors, axis=1)
    similarities = np.dot(document_vectors, query_vector) / (norms_docs * norms_query)

    # Initialize set of selected indices and results list
    selected_indices = set()
    result_indices = []

    # Diversified search loop
    for _ in range(top_n):
        max_mmr = float("-inf")
        best_index = None

        # Loop over all documents
        for idx, _doc_vector in enumerate(document_vectors):
            if idx not in selected_indices:
                relevance = relevance_factor * similarities[idx]
                diversity = 0

                # Penalize based on similarity to already selected documents
                if selected_indices:
                    max_sim_to_selected = max(
                        [
                            np.dot(document_vectors[idx], document_vectors[s])
                            / (
                                np.linalg.norm(document_vectors[idx])
                                * np.linalg.norm(document_vectors[s])
                            )
                            for s in selected_indices
                        ]
                    )
                    diversity = (1 - relevance_factor) * max_sim_to_selected

                mmr_score = relevance - diversity

                # Update best MMR score and index
                if mmr_score > max_mmr:
                    max_mmr = mmr_score
                    best_index = idx

        # Add the best document to the result and mark it as selected
        if best_index is None:
            logger.warning(f"No best index found over {len(document_vectors)} documents.")
            continue
        result_indices.append(best_index)
        selected_indices.add(best_index)

    return result_indices



"""Chat with a KB."""

import logging
import time
from dataclasses import dataclass
from typing import ClassVar, Dict, List, Optional

import requests

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "http://wwwdev.ebi.ac.uk/Tools/omicsdi/ws"


@dataclass
class OmicsDIWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over OMICS DI Search.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "omicsdi"

    default_object_type = "Dataset"

    source: str = None  # pride, ...

    def external_search(
        self, text: str, expand: bool = False, where: Optional[Dict] = None, **kwargs
    ) -> List[Dict]:
        """
        Search the OmicsDI database for the given text.

        TODO: full text retrieval doesn't seem to work well
        """
        params = {
            "size": 20,
            "faceCount": 30,
        }
        if expand:

            def qt(t: str):
                t = t.strip()
                if " " in t:
                    return f'"{t}"'
                return t

            logger.info(f"Expanding search term: {text} to create OmicsDI query")
            model = self.extractor.model
            response = model.prompt(
                text,
                system="""
                Take the specified search text, and expand it to a list
                of key terms used to construct a query. You will return results as
                semi-colon separated list of the most relevant terms. Make sure to
                include all relevant concepts in the returned terms.""",
            )
            terms = response.text().split(";")
            logger.info(f"Expanded terms: {terms}")
            terms = [qt(t) for t in terms]
            # terms = terms[0:2]
            search_term = " ".join(terms)
        else:
            search_term = text
        params["query"] = search_term
        if where:
            params.update(where)
        time.sleep(0.25)
        logger.info(f"Constructed query: {params}")
        url = f"{BASE_URL}/dataset/search"
        logger.info(f"Searching OmicsDI {url} with query: {params}")
        response = requests.get(url, params=params)
        data = response.json()

        datasets = data["datasets"]
        logger.info(f"Found {len(datasets)} datasets")
        for dataset in datasets:
            dataset["additional"] = self.additional_info(dataset["id"], dataset["source"])
        return datasets

    def additional_info(self, local_id: str, source: str):
        """
        Augment the local ID with information from the database.
        """
        url = f"{BASE_URL}/dataset/{source}/{local_id}"
        logger.info(f"Getting additional info from {url}")
        response = requests.get(url)
        response.raise_for_status()
        data = response.json()
        return data["additional"]

    def objects_by_ids(self, ids: List[str], source: str = None):
        """
        Augment the local ID with information from the database.
        """
        source = source or self.source
        data_objects = []
        for id in ids:
            if ":" in id:
                source, local_id = id.split(":")
            else:
                local_id = id
            if not source:
                raise ValueError(f"Need a source for ID {id}")
            source = source.lower()
            url = f"{BASE_URL}/dataset/{source}/{local_id}"
            logger.info(f"Getting additional info from {url}")
            response = requests.get(url)
            response.raise_for_status()
            data = response.json()
            data_objects.append(data)

        return data_objects



"""
Wrappers on top of external data sources.

Wrappers allow for dynamic or static loading of an external data source
into a store.
"""

from curategpt.wrappers.base_wrapper import BaseWrapper

__all__ = [
    "BaseWrapper",
    "PubmedWrapper",
    "WikipediaWrapper",
    "BiocWrapper",
    "OntologyWrapper",
    "BioportalWrapper",
    "NMDCWrapper",
    "GOCAMWrapper",
    "JGIWrapper",
    "AllianceGeneWrapper",
    "NCBIBiosampleWrapper",
    "NCBIBioprojectWrapper",
    "OmicsDIWrapper",
    "UniprotWrapper",
    "OBOFormatWrapper",
    "ClinVarWrapper",
    "PMCWrapper",
    "HPOAWrapper",
    "HPOAByPubWrapper",
    "MAXOAWrapper",
    "GoogleDriveWrapper",
    "FAIRSharingWrapper",
    "FilesystemWrapper",
    "get_wrapper",
]


def get_all_subclasses(cls):
    """Recursively get all subclasses of a given class."""
    direct_subclasses = cls.__subclasses__()
    return direct_subclasses + [
        s for subclass in direct_subclasses for s in get_all_subclasses(subclass)
    ]


def get_wrapper(name: str, **kwargs) -> BaseWrapper:
    # NOTE: ORDER DEPENDENT. TODO: fix this
    from curategpt.wrappers.bio.alliance_gene_wrapper import \
        AllianceGeneWrapper  # noqa
    from curategpt.wrappers.bio.bacdive_wrapper import BacDiveWrapper  # noqa
    from curategpt.wrappers.bio.gocam_wrapper import GOCAMWrapper  # noqa
    from curategpt.wrappers.bio.mediadive_wrapper import \
        MediaDiveWrapper  # noqa
    from curategpt.wrappers.bio.omicsdi_wrapper import OmicsDIWrapper  # noqa
    from curategpt.wrappers.bio.reactome_wrapper import \
        ReactomeWrapper  # noqa
    from curategpt.wrappers.bio.uniprot_wrapper import UniprotWrapper  # noqa
    from curategpt.wrappers.clinical.clinvar_wrapper import \
        ClinVarWrapper  # noqa
    from curategpt.wrappers.clinical.ctgov_wrapper import \
        ClinicalTrialsWrapper  # noqa
    from curategpt.wrappers.clinical.hpoa_by_pub_wrapper import \
        HPOAByPubWrapper  # noqa
    from curategpt.wrappers.clinical.hpoa_wrapper import HPOAWrapper  # noqa
    from curategpt.wrappers.clinical.maxoa_wrapper import MAXOAWrapper  # noqa
    from curategpt.wrappers.general.filesystem_wrapper import \
        FilesystemWrapper  # noqa
    from curategpt.wrappers.general.github_wrapper import \
        GitHubWrapper  # noqa
    from curategpt.wrappers.general.google_drive_wrapper import \
        GoogleDriveWrapper  # noqa
    from curategpt.wrappers.general.gspread_wrapper import \
        GSpreadWrapper  # noqa
    from curategpt.wrappers.general.json_wrapper import JSONWrapper  # noqa
    from curategpt.wrappers.general.linkml_schema_wrapper import \
        LinkMLSchemarapper  # noqa
    from curategpt.wrappers.investigation.ess_deepdive_wrapper import \
        ESSDeepDiveWrapper  # noqa
    from curategpt.wrappers.investigation.fairsharing_wrapper import \
        FAIRSharingWrapper  # noqa
    from curategpt.wrappers.investigation.jgi_wrapper import \
        JGIWrapper  # noqa
    from curategpt.wrappers.investigation.ncbi_bioproject_wrapper import \
        NCBIBioprojectWrapper  # noqa
    from curategpt.wrappers.investigation.ncbi_biosample_wrapper import \
        NCBIBiosampleWrapper  # noqa
    from curategpt.wrappers.investigation.nmdc_wrapper import \
        NMDCWrapper  # noqa
    from curategpt.wrappers.legal.reusabledata_wrapper import \
        ReusableDataWrapper  # noqa
    from curategpt.wrappers.literature.bioc_wrapper import BiocWrapper  # noqa
    from curategpt.wrappers.literature.pmc_wrapper import PMCWrapper  # noqa
    from curategpt.wrappers.literature.pubmed_wrapper import \
        PubmedWrapper  # noqa
    from curategpt.wrappers.literature.wikipedia_wrapper import \
        WikipediaWrapper  # noqa
    from curategpt.wrappers.ontology.bioportal_wrapper import \
        BioportalWrapper  # noqa
    from curategpt.wrappers.ontology.oboformat_wrapper import \
        OBOFormatWrapper  # noqa
    from curategpt.wrappers.ontology.ontology_wrapper import \
        OntologyWrapper  # noqa

    for c in get_all_subclasses(BaseWrapper):
        if c.name == name:
            return c(**kwargs)
    raise ValueError(f"Unknown view {name}, not found in {get_all_subclasses(BaseWrapper)}")



from typing import List

from pydantic import BaseModel, ConfigDict, Extra


class Relationship(BaseModel):
    """
    A relationship to another node.

    Corresponds to an edge in an OBO graph.
    """

    model_config = ConfigDict(protected_namespaces=())
    predicate: str
    target: str


class OntologyClass(BaseModel, extra=Extra.allow):
    """
    An ontology class.

    Corresponds to a node in an OBO graph.
    """

    model_config = ConfigDict(protected_namespaces=())
    id: str
    label: str = None
    definition: str = None
    aliases: List[str] = None
    relationships: List[Relationship] = None
    logical_definition: List[Relationship] = None
    original_id: str = None


class Ontology(BaseModel):
    """
    An ontology.

    Corresponds to an OBO graph.
    """

    model_config = ConfigDict(protected_namespaces=())
    elements: List[OntologyClass] = None



import csv
import logging
from dataclasses import dataclass, field
from typing import List, TextIO

import yaml

from curategpt.agents.dragon_agent import DragonAgent
from curategpt.evaluation.base_evaluator import BaseEvaluator
from curategpt.evaluation.calc_statistics import (
    aggregate_metrics,
    calculate_metrics,
    evaluate_predictions,
)
from curategpt.evaluation.evaluation_datamodel import ClassificationMetrics, ClassificationOutcome

logger = logging.getLogger(__name__)


@dataclass
class DatabaseAugmentedCompletionEvaluator(BaseEvaluator):
    """
    Retrieves objects in response to a query using a structured knowledge source.
    """

    agent: DragonAgent = None
    fields_to_predict: List[str] = field(default_factory=list)
    fields_to_mask: List[str] = field(default_factory=list)

    def evaluate(
        self,
        test_collection: str,
        num_tests: int = None,
        report_file: TextIO = None,
        report_tsv_file: TextIO = None,
        working_directory: str = None,
        **kwargs,
    ) -> ClassificationMetrics:
        """
        Evaluate the agent on a test collection.

        Note: the main collection used for few-shot learning is passed in kwargs (this may change)

        :param test_collection:
        :param num_tests:
        :param report_file:
        :param kwargs:
        :return:
        """
        agent = self.agent
        db = agent.knowledge_source
        test_objs = list(db.find(collection=test_collection))
        if any(obj for obj in test_objs if any(f not in obj for f in self.fields_to_predict)):
            logger.info("Alternate strategy to get test objs; query whole collection")
            test_objs = db.peek(collection=test_collection, limit=1000000)
            test_objs = [obj for obj in test_objs if all(f in obj for f in self.fields_to_predict)]
            test_objs = test_objs[:num_tests]
        if num_tests and len(test_objs) < num_tests:
            raise ValueError(
                f"Insufficient test objects in collection {test_collection}; "
                f"{len(test_objs)} < {num_tests}"
            )
        all_metrics = []
        if not test_objs:
            raise ValueError(f"No test objects found in collection {test_collection}")
        n = 0
        results_dictwriter = None
        field_names = []
        for test_obj in test_objs:
            field_names.extend([k for k in test_obj.keys() if k not in field_names])
        for test_obj in test_objs:
            test_obj_query = {
                k: v
                for k, v in test_obj.items()
                if k not in self.fields_to_predict and k not in self.fields_to_mask
            }
            logger.debug(f"## Query: {test_obj_query}")
            ao = agent.complete(
                test_obj_query,
                fields_to_predict=self.fields_to_predict,
                fields_to_mask=self.fields_to_mask,
                **kwargs,
            )
            logger.debug(f"--- Expected: {test_obj}")
            logger.debug(f"--- Prediction: {ao.object}")
            outcomes = []
            for f in self.fields_to_predict:
                outcomes.extend(
                    list(evaluate_predictions(ao.object.get(f, None), test_obj.get(f, None)))
                )
            for outcome, info in outcomes:
                if outcome != ClassificationOutcome.TRUE_POSITIVE:
                    logger.debug(f"## Diff: {outcome} - {info}")
            metrics = calculate_metrics(outcomes)
            all_metrics.append(metrics)
            logger.info(f"## Metrics: {metrics.json()}")
            aggregated = aggregate_metrics(all_metrics)
            logger.info(f"## Aggregated: {aggregated.json()}")
            n += 1
            if report_file:
                report_file.write(f"# RESULTS {n}\n")
                report_file.write(f"## Query:\n{yaml.dump(test_obj_query)}\n---\n")
                report_file.write(f"## Prompt:\n{ao.annotations.get('prompt', None)}\n---\n")
                report_file.write(f"## Expected:\n{yaml.dump(test_obj)}\n---\n")
                report_file.write(f"## Predicted:\n{yaml.dump(ao.object)}\n---\n")
                report_file.write(f"## Metrics:\n{yaml.dump(metrics.dict())}\n---\n")
                report_file.write(f"## Aggregated:\n{yaml.dump(aggregated.dict())}\n---\n")
            if report_tsv_file:
                row = {}
                for k, v in test_obj.items():
                    if k in self.fields_to_predict:
                        row[f"expected_{k}"] = v
                    elif k in self.fields_to_mask:
                        row[f"masked_{k}"] = v
                    else:
                        row[f"feature_{k}"] = v
                for f in self.fields_to_predict:
                    row[f"predicted_{f}"] = ao.object.get(f, None)
                    # row[f"expected_{f}"] = test_obj.get(f, None)
                for k, v in metrics.dict().items():
                    row[f"metric_{k}"] = v
                if n == 1:
                    field_names.extend([k for k in row.keys() if k not in field_names])
                    results_dictwriter = csv.DictWriter(
                        report_tsv_file, fieldnames=field_names, delimiter="\t"
                    )
                    results_dictwriter.writeheader()
                if not results_dictwriter:
                    raise AssertionError("results_dictwriter not initialized")
                results_dictwriter.writerow({k: v for k, v in row.items() if k in field_names})
                report_tsv_file.flush()
        aggregated = aggregate_metrics(all_metrics)
        return aggregated

    def evaluate_object(self, obj, **kwargs) -> ClassificationMetrics:
        pass



"""
This is a DuckDB adapter for the Vector Similarity Search (VSS) extension
using the experimental persistence feature
"""

import json
import logging
import os
import re
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Dict, Iterable, Iterator, List, Mapping, Optional, Union

import duckdb
import llm
import numpy as np
import openai
import psutil
import yaml
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.yamlutils import YAMLRoot
from oaklib.utilities.iterator_utils import chunk
from openai import OpenAI
from pydantic import BaseModel
from sentence_transformers import SentenceTransformer

from curategpt.store.db_adapter import DBAdapter
from curategpt.store.duckdb_result import DuckDBSearchResult
from curategpt.store.metadata import CollectionMetadata
from curategpt.store.vocab import (
    DEFAULT_MODEL,
    DEFAULT_OPENAI_MODEL,
    DISTANCES,
    DOCUMENTS,
    EMBEDDINGS,
    IDS,
    METADATAS,
    MODEL_MAP,
    OBJECT,
    PROJECTION,
    QUERY,
    SEARCH_RESULT,
)
from curategpt.utils.vector_algorithms import mmr_diversified_search

logger = logging.getLogger(__name__)


@dataclass
class DuckDBAdapter(DBAdapter):
    name: ClassVar[str] = "duckdb"
    default_model: str = "all-MiniLM-L6-v2"
    conn: duckdb.DuckDBPyConnection = field(init=False)
    vec_dimension: int = field(init=False)
    ef_construction: int = 128
    ef_search: int = 64
    M: int = 16
    distance_metric: str = "cosine"
    id_field: str = "id"
    text_lookup: Optional[Union[str, Callable]] = field(default="text")
    id_to_object: Mapping[str, dict] = field(default_factory=dict)
    default_max_document_length: ClassVar[int] = 6000
    openai_client: OpenAI = field(default=None)

    def __post_init__(self):
        if not self.path:
            self.path = "./db/db_file.duckdb"
        if os.path.isdir(self.path):
            self.path = os.path.join("./db", self.path, "db_file.duckdb")
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            logger.info(
                f"Path {self.path} is a directory. Using {self.path} as the database path\n\
            as duckdb needs a file path"
            )
        self.ef_construction = self._validate_ef_construction(self.ef_construction)
        self.ef_search = self._validate_ef_search(self.ef_search)
        self.M = self._validate_m(self.M)
        logger.info(f"Using DuckDB at {self.path}")
        # handling concurrency
        try:
            self.conn = duckdb.connect(self.path, read_only=False)
        except duckdb.IOException as e:
            match = re.search(r"PID (\d+)", str(e))
            if match:
                pid = int(match.group(1))
                logger.info(f"Got {e}.Attempting to kill process with PID: {pid}")
                self.kill_process(pid)
                self.conn = duckdb.connect(self.path, read_only=False)
            else:
                logger.error(f"{e} without PID information.")
                raise
        self.conn.execute("INSTALL vss;")
        self.conn.execute("LOAD vss;")
        self.conn.execute("SET hnsw_enable_experimental_persistence=true;")
        if self.default_model is None:
            self.model = self.default_model
        self.vec_dimension = self._get_embedding_dimension(self.default_model)

    def _initialize_openai_client(self):
        if self.openai_client is None:
            from dotenv import load_dotenv

            load_dotenv()
            openai_api_key = os.environ.get("OPENAI_API_KEY")
            if openai_api_key:
                self.openai_client = openai.OpenAI(api_key=openai_api_key)
            else:
                raise openai.OpenAIError(
                    "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
                )

    def _get_collection_name(self, collection: Optional[str] = None) -> str:
        """
        Get the collection name or the default collection name
        :param collection:
        :return:
        """
        return self._get_collection(collection)

    def _create_table_if_not_exists(
        self, collection: str, vec_dimension: int, distance: str, model: str = None
    ):
        """
        Create a table for the given collection if it does not exist
        :param collection:
        :return:
        """
        logger.info(
            f"Table {collection} does not exist, creating ...: PARAMS: model: {model}, distance: {distance},\
        vec_dimension: {vec_dimension}"
        )
        if model is None:
            model = self.default_model
            logger.info(f"Model in create_table_if_not_exists: {model}")
        if distance is None:
            distance = self.distance_metric
        safe_collection_name = f'"{collection}"'
        create_table_sql = f"""
            CREATE TABLE IF NOT EXISTS {safe_collection_name} (
                id VARCHAR PRIMARY KEY,
                metadata JSON,
                embeddings FLOAT[{vec_dimension}],
                documents TEXT
            )
        """
        self.conn.execute(create_table_sql)

        metadata = CollectionMetadata(name=collection, model=model, hnsw_space=distance)
        metadata_json = json.dumps(metadata.dict(exclude_none=True))
        safe_collection_name = f'"{collection}"'
        self.conn.execute(
            f"""
                    INSERT INTO {safe_collection_name} (id, metadata) VALUES ('__metadata__', ?)
                    ON CONFLICT (id) DO NOTHING
                    """,
            [metadata_json],
        )

    def create_index(self, collection: str):
        """
        Create an index for the given collection
        Parameters
        ----------
        collection

        Returns
        -------

        """
        cm = self.collection_metadata(collection)
        safe_collection_name = f'"{collection}"'
        index_name = f"{collection}_index"
        create_index_sql = f"""
            CREATE INDEX IF NOT EXISTS "{index_name}" ON {safe_collection_name}
            USING HNSW (embeddings) WITH (
                metric='{cm.hnsw_space}',
                ef_construction={self.ef_construction},
                ef_search={self.ef_search},
                M={self.M}
            )
        """
        self.conn.execute(create_index_sql)

    def _embedding_function(
        self, texts: Union[str, List[str], List[List[str]]], model: str = None
    ) -> list:
        """
        Get the embeddings for the given texts using the specified model
        :param texts: A single text or a list of texts to embed
        :param model: Model to use for embedding
        :return: A single embedding or a list of embeddings
        """
        single_text = False
        if isinstance(texts, str):
            texts = [texts]
            single_text = True

        if model is None:
            model = self.model

        if model.startswith("openai:"):
            self._initialize_openai_client()
            openai_model = model.split(":", 1)[1]
            if openai_model == "" or openai_model not in MODEL_MAP.keys():
                logger.info(
                    f"The model {openai_model} is not "
                    f"one of {[MODEL_MAP.keys()]}. Defaulting to {DEFAULT_OPENAI_MODEL}"
                )
                openai_model = DEFAULT_OPENAI_MODEL

            responses = [
                self.openai_client.embeddings.create(input=text, model=openai_model)
                .data[0]
                .embedding
                for text in texts
            ]
            return responses[0] if single_text else responses

        model = SentenceTransformer(model)
        embeddings = model.encode(texts, convert_to_tensor=False).tolist()
        return embeddings[0] if single_text else embeddings

    def insert(self, objs: Union[OBJECT, Iterable[OBJECT]], **kwargs):
        """
        Insert objects into the collection
        :param objs:
        :param kwargs:
        :return:
        """
        logger.info(f"\n\nIn insert duckdb, {kwargs.get('model')}\n\n")
        self._process_objects(objs, method="insert", **kwargs)

    # DELETE first to ensure primary key  constraint https://duckdb.org/docs/sql/indexes
    def update(self, objs: Union[OBJECT, Iterable[OBJECT]], **kwargs):
        """
        Update objects in the collection.
        :param objs:
        :param kwargs:
        :return:
        """
        collection = kwargs.get("collection")
        ids = [self._id(o, self.id_field) for o in objs]
        safe_collection_name = f'"{collection}"'
        delete_sql = f"DELETE FROM {safe_collection_name} WHERE id = ?"
        logger.info("DELETED collection: {collection}")
        self.conn.executemany(delete_sql, [(id_,) for id_ in ids])
        logger.info(f"INSERTING collection: {collection}")
        self.insert(objs, **kwargs)

    def upsert(self, objs: Union[OBJECT, Iterable[OBJECT]], **kwargs):
        """
        Upsert objects into the collection
        :param objs:
        :param kwargs:
        :return:
        """
        collection = kwargs.get("collection")
        logger.info(f"\n\nUpserting objects into collection {collection}\n\n")
        logger.info(f"model in upsert: {kwargs.get('model')}, distance: {self.distance_metric}")
        if collection not in self.list_collection_names():
            vec_dimension = self._get_embedding_dimension(kwargs.get("model"))
            self._create_table_if_not_exists(
                collection, vec_dimension, model=kwargs.get("model"), distance=self.distance_metric
            )
        ids = [self._id(o, self.id_field) for o in objs]
        existing_ids = set()
        for id_ in ids:
            safe_collection_name = f'"{collection}"'
            result = self.conn.execute(
                f"SELECT id FROM {safe_collection_name} WHERE id = ?", [id_]
            ).fetchall()
            if result:
                existing_ids.add(id_)
        objs_to_update = [o for o in objs if self._id(o, self.id_field) in existing_ids]
        objs_to_insert = [o for o in objs if self._id(o, self.id_field) not in existing_ids]
        if objs_to_update:
            logger.info(f"in Upsert and updating now in collection: {collection}")
            self.update(objs_to_update, **kwargs)

        if objs_to_insert:
            logger.info(f"in Upsert and inserting now in collection: {collection}")
            self.insert(objs_to_insert, **kwargs)

    def _process_objects(
        self,
        objs: Union[OBJECT, Iterable[OBJECT]],
        collection: str = None,
        batch_size: int = None,
        object_type: str = None,
        model: str = None,
        distance: str = None,
        text_field: Union[str, Callable] = None,
        method: str = "insert",
        **kwargs,
    ):
        """
        Process objects by inserting, updating or upserting them into the collection
        :param objs:
        :param collection:
        :param batch_size:
        :param object_type:
        :param model:
        :param text_field:
        :param method:
        :param kwargs:
        :return:
        """
        collection = self._get_collection_name(collection)
        logger.info(f"Processing objects for collection {collection}")
        self.vec_dimension = self._get_embedding_dimension(model)
        logger.info(f"(process_objects: Model: {model}, vec_dimension: {self.vec_dimension}")
        if collection not in self.list_collection_names():
            logger.info(f"(process)Creating table for collection {collection}")
            self._create_table_if_not_exists(
                collection, self.vec_dimension, model=model, distance=distance
            )
        if isinstance(objs, Iterable) and not isinstance(objs, str):
            objs = list(objs)
        else:
            objs = [objs]
        obj_count = len(objs)
        kwargs.update({"object_count": obj_count})
        cm = self.collection_metadata(collection)
        if batch_size is None:
            batch_size = 100000
        if text_field is None:
            text_field = self.text_lookup
        id_field = self.id_field
        sql_command = self._generate_sql_command(collection, method)
        sql_command = sql_command.format(collection=collection)
        if not self._is_openai(collection):
            for next_objs in chunk(objs, batch_size):
                next_objs = list(next_objs)
                docs = [self._text(o, text_field) for o in next_objs]
                metadatas = [self._dict(o) for o in next_objs]
                ids = [self._id(o, id_field) for o in next_objs]
                embeddings = self._embedding_function(docs, cm.model)
                try:
                    self.conn.execute("BEGIN TRANSACTION;")
                    self.conn.executemany(
                        sql_command, list(zip(ids, metadatas, embeddings, docs))  # noqa: B905
                    )
                    # reason to block B905: codequality check
                    # blocking 3.11 because only code quality issue and 3.9 gives value error with keyword strict
                    # TODO: delete after PR#76 is merged
                    self.conn.execute("COMMIT;")
                except Exception as e:
                    self.conn.execute("ROLLBACK;")
                    logger.error(
                        f"Transaction failed: {e}, default model: {self.default_model}, model used: {model}, len(embeddings): {len(embeddings[0])}"
                    )
                    raise
                finally:
                    self.create_index(collection)
        else:
            if model.startswith("openai:"):
                openai_model = model.split(":", 1)[1]
                if openai_model == "" or openai_model not in MODEL_MAP.keys():
                    logger.info(
                        f"The model {openai_model} is not "
                        f"one of {MODEL_MAP.keys()}. Defaulting to {DEFAULT_OPENAI_MODEL}"
                    )
                    openai_model = DEFAULT_OPENAI_MODEL  # ada 002
                else:
                    logger.error(f"Something went wonky ## model: {model}")
            from transformers import GPT2Tokenizer

            tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
            for next_objs in chunk(objs, batch_size):  # Existing chunking
                next_objs = list(next_objs)
                docs = [self._text(o, text_field) for o in next_objs]
                metadatas = [self._dict(o) for o in next_objs]
                ids = [self._id(o, id_field) for o in next_objs]

                tokenized_docs = [tokenizer.encode(doc) for doc in docs]
                current_batch = []
                current_token_count = 0
                batch_embeddings = []

                i = 0
                while i < len(tokenized_docs):
                    doc_tokens = tokenized_docs[i]
                    # peek
                    if current_token_count + len(doc_tokens) <= 8192:
                        current_batch.append(doc_tokens)
                        current_token_count += len(doc_tokens)
                        i += 1
                    else:
                        if current_batch:
                            logger.info(f"Tokens: {current_token_count}")
                            texts = [tokenizer.decode(tokens) for tokens in current_batch]
                            short_name, _ = MODEL_MAP[openai_model]
                            embedding_model = llm.get_embedding_model(short_name)
                            logger.info(f"Number of texts/docs to embed in batch: {len(texts)}")
                            embeddings = list(embedding_model.embed_multi(texts, len(texts)))
                            logger.info(f"Number of Documents in batch: {len(embeddings)}")
                            batch_embeddings.extend(embeddings)

                        if len(doc_tokens) > 8192:
                            logger.warning(
                                f"Document with ID {ids[i]} exceeds the token limit alone and will be skipped."
                            )
                            # try:
                            #     embeddings = OpenAIEmbeddings(model=model, tiktoken_model_name=model).embed_query(texts,
                            #     embeddings.average                                                                                model)
                            #     batch_embeddings.extend(embeddings)
                            # skipping
                            i += 1
                            continue
                        else:
                            current_batch = []
                            current_token_count = 0

                if current_batch:
                    logger.info(f"Last batch, token count: {current_token_count}")
                    texts = [tokenizer.decode(tokens) for tokens in current_batch]
                    short_name, _ = MODEL_MAP[openai_model]
                    embedding_model = llm.get_embedding_model(short_name)
                    embeddings = list(embedding_model.embed_multi(texts))
                    batch_embeddings.extend(embeddings)
                logger.info(
                    f"Trying to insert: {len(ids)} IDS, {len(metadatas)} METADATAS, {len(batch_embeddings)} EMBEDDINGS"
                )
                try:
                    self.conn.execute("BEGIN TRANSACTION;")
                    self.conn.executemany(
                        sql_command, list(zip(ids, metadatas, batch_embeddings, docs, strict=False))
                    )
                    self.conn.execute("COMMIT;")
                except Exception as e:
                    self.conn.execute("ROLLBACK;")
                    logger.error(
                        f"Transaction failed: {e}, default model: {self.default_model}, model used: {model}, len(embeddings): {len(embeddings[0])}"
                    )
                    raise
                finally:
                    self.create_index(collection)

    def remove_collection(self, collection: str = None, exists_ok=False, **kwargs):
        """
        Remove the collection from the database
        :param collection:
        :param exists_ok:
        :param kwargs:
        :return:
        """
        collection = self._get_collection(collection)
        if not exists_ok:
            if collection not in self.list_collection_names():
                raise ValueError(f"Collection {collection} does not exist")
        # duckdb, requires that identifiers containing special characters ("-") must be enclosed in double quotes.
        safe_collection_name = f'"{collection}"'
        self.conn.execute(f"DROP TABLE IF EXISTS {safe_collection_name}")

    def search(
        self,
        text: str,
        where: QUERY = None,
        collection: str = None,
        limit: int = 10,
        relevance_factor: float = None,
        include=None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        """
        Search for objects in the collection that match the given text
        :param text:
        :param where:
        :param collection:
        :param limit:
        :param relevance_factor:
        :param include:
        :param kwargs:
        :return:
        """
        yield from self._search(
            text=text,
            where=where,
            collection=collection,
            limit=limit,
            relevance_factor=relevance_factor,
            include=include,
            **kwargs,
        )

    def _search(
        self,
        text: str,
        where: QUERY = None,
        collection: str = None,
        limit: int = 10,
        relevance_factor: float = None,
        model: str = None,
        include=None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        if relevance_factor is not None and relevance_factor < 1.0:
            yield from self._diversified_search(
                text=text,
                where=where,
                collection=collection,
                limit=limit,
                include=include,
                relevance_factor=relevance_factor,
                **kwargs,
            )
            return
        if include is None:
            include = {METADATAS, DOCUMENTS, DISTANCES}
        else:
            include = set(include)
        collection = self._get_collection(collection)
        cm = self.collection_metadata(collection)
        logger.info(f"Collection metadata={cm}")
        if model is None:
            if cm:
                model = cm.model
            if model is None:
                model = self.default_model
        logger.info(f"Model={model}")
        where_conditions = []
        if where:
            where_conditions.append(where)
        where_clause = " AND ".join(where_conditions)
        if where_clause:
            where_clause = f"WHERE {where_clause}"
        if relevance_factor is not None and relevance_factor < 1.0:
            yield from self._diversified_search(
                text, where, collection, limit, relevance_factor, include, **kwargs
            )
            return
        query_embedding = self._embedding_function(text, model)
        safe_collection_name = f'"{collection}"'

        vec_dimension = self._get_embedding_dimension(model)

        # TODO: !VERY IMPORTANT! distance metrics between Chroma and DuckDB have very different, unclear implementations
        # https://duckdb.org/docs/sql/functions/array.html#array_distancearray1-array2
        # https://docs.trychroma.com/guides
        # chromaDB: by default l2, other options are ip and cosine
        # duckDB: by default none, array_distance() or 1-array_cosine_similarity(), both bring different distances
        # than chromaDBs distance metric
        results = self.conn.execute(
            f"""
            SELECT *, array_distance(embeddings::FLOAT[{vec_dimension}],
            {query_embedding}::FLOAT[{vec_dimension}]) as distance
            FROM {safe_collection_name}
            {where_clause}
            ORDER BY distance
            LIMIT ?
        """,
            [limit],
        ).fetchall()
        yield from self.parse_duckdb_result(results, include)

    def _diversified_search(
        self,
        text: str,
        where: QUERY = None,
        collection: str = None,
        limit: int = 10,
        relevance_factor: float = 0.5,
        include=None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        if limit is None:
            limit = 10
        # we need to set this as we need EMBEDDINGS
        include = {METADATAS, DOCUMENTS, EMBEDDINGS, DISTANCES}
        collection = self._get_collection(collection)
        cm = self.collection_metadata(collection)
        where_conditions = []
        if where:
            for key, value in where.items():
                where_conditions.append(f"json_extract(metadata, '$.{key}') = '{value}'")
        where_clause = " AND ".join(where_conditions)
        if where_clause:
            where_clause = f"WHERE {where_clause}"
        query_embedding = self._embedding_function(text, model=cm.model)
        safe_collection_name = f'"{collection}"'
        vec_dimension = self._get_embedding_dimension(cm.model)
        results = self.conn.execute(
            f"""
                    SELECT *, array_distance(embeddings::FLOAT[{vec_dimension}],
                    {query_embedding}::FLOAT[{vec_dimension}]) as distance
                    FROM {safe_collection_name}
                    {where_clause}
                    ORDER BY distance
                    LIMIT ?
                """,
            [limit * 10],
        ).fetchall()
        results = list(self.parse_duckdb_result(results, include))
        if not results:
            return
        rows = [np.array(r[2]["_embeddings"]) for r in results]
        query = np.array(query_embedding)
        reranked_indices = mmr_diversified_search(
            query, rows, relevance_factor=relevance_factor, top_n=limit
        )
        for i in reranked_indices:
            yield results[i]

    def list_collection_names(self):
        """
        List the names of all collections in the database
        :return:
        """
        result = self.conn.execute("PRAGMA show_tables;").fetchall()
        return [row[0] for row in result]

    def collection_metadata(
        self, collection_name: Optional[str] = None, include_derived=False, **kwargs
    ) -> Optional[CollectionMetadata]:
        """
        Get the metadata for the collection
        :param collection_name:
        :param include_derived:
        :param kwargs:
        :return:
        """
        collection_name = self._get_collection(collection_name)
        safe_collection_name = f'"{collection_name}"'
        try:
            result = self.conn.execute(
                f"SELECT metadata FROM {safe_collection_name} WHERE id = '__metadata__'"
            ).fetchone()
            if result:
                metadata = json.loads(result[0])
                metadata_instance = CollectionMetadata(**metadata)
                if include_derived:
                    # not implemented yet
                    # metadata_instance.object_count = compute_object_count(collection_name
                    pass
                return metadata_instance
        except Exception as e:
            logger.error(f"Failed to retrieve metadata for collection {collection_name}: {str(e)}")
            return None

    def update_collection_metadata(self, collection: str, **kwargs):
        """
        Update the metadata for a collection. This function will merge new metadata provided
        via kwargs with existing metadata, if any, ensuring that only the specified fields
        are updated.
        :param collection:
        :param kwargs:
        :return:
        """
        if not collection:
            raise ValueError("Collection name must be provided.")
        current_metadata = self.collection_metadata(collection)
        if current_metadata is None:
            current_metadata = CollectionMetadata(**kwargs)
        else:
            for key, value in kwargs.items():
                if hasattr(current_metadata, key):
                    setattr(current_metadata, key, value)
        metadata_dict = current_metadata.dict(exclude_none=True)
        metadata_json = json.dumps(metadata_dict)
        safe_collection_name = f'"{collection}"'
        self.conn.execute(
            f"""
                UPDATE {safe_collection_name} SET metadata = ?
                WHERE id = '__metadata__'
                """,
            [metadata_json],
        )
        return current_metadata

    def set_collection_metadata(
        self, collection_name: Optional[str], metadata: CollectionMetadata, **kwargs
    ):
        """
        Set the metadata for the collection
        :param collection_name:
        :param metadata:
        :param kwargs:
        :return:
        """
        if collection_name is None:
            raise ValueError("Collection name must be provided.")

        metadata_json = json.dumps(metadata.dict(exclude_none=True))
        safe_collection_name = f'"{collection_name}"'
        self.conn.execute(
            f"""
            UPDATE {safe_collection_name}
            SET metadata = ?
            WHERE id = '__metadata__'
            """,
            [metadata_json],
        )

    def find(
        self,
        where: QUERY = None,
        projection: PROJECTION = None,
        collection: str = None,
        include=None,
        limit: int = 10,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        """
        Find objects in the collection that match the given query and projection

        :param where: the query to filter the results
        :param projection:
        :param collection: name of the collection to search
        :param include: fields to be included in output
        :param limit: maximum number of results to return
        :param kwargs:
        :return:

        Parameters
        ----------
        """
        collection = self._get_collection(collection)
        where_clause = self._parse_where_clause(where) if where else ""
        where_clause = f"WHERE {where_clause}" if where_clause else ""
        if include is None:
            include = [IDS, METADATAS, DOCUMENTS]
        safe_collection_name = f'"{collection}"'
        query = f"""
                    SELECT id, metadata, embeddings, documents, NULL as distance
                    FROM {safe_collection_name}
                    {where_clause}
                    LIMIT {limit}
                """
        results = self.conn.execute(query).fetchall()
        yield from self.parse_duckdb_result(results, include)

    def matches(self, obj: OBJECT, include=None, **kwargs) -> Iterator[SEARCH_RESULT]:
        """
        Find objects in the collection that match the given object
        :param obj:
        :param include:
        :param kwargs:
        :return:
        """
        if include is None:
            include = {IDS, METADATAS, DOCUMENTS, DISTANCES}
        else:
            include = set(include)
        text_field = self.text_lookup
        logger.info(f"## TEXT FIELD:{text_field}")
        text = self._text(obj, text_field)
        logger.info(f"{text}")
        logger.info(f"Query term: {text}")
        yield from self.search(text=text, include=include, **kwargs)

    def lookup(self, id: str, collection: str = None, include=None, **kwargs) -> OBJECT:
        """
        Lookup an object by its id
        :param id: ID of the object to lookup
        :param collection: Name of the collection to search
        :param include: List of fields to include in the output ['metadata', 'embeddings', 'documents']
        :param kwargs:
        :return:
        """
        if include is None:
            include = {METADATAS}
        else:
            include = set(include)
        safe_collection_name = f'"{collection}"'
        result = self.conn.execute(
            f"""
                SELECT *
                FROM {safe_collection_name}
                WHERE id = ?
            """,
            [id],
        ).fetchone()
        if isinstance(result, tuple) and len(result) > 1:
            search_result = DuckDBSearchResult(
                ids=result[0],
                metadatas=json.loads(result[1]),
                embeddings=result[2],
                documents=result[3],
                include=include,
            )
            return search_result.to_dict().get(METADATAS)

    def peek(
        self, collection: str = None, limit=5, include=None, offset: int = 0, **kwargs
    ) -> Iterator[SEARCH_RESULT]:
        """
        Peek at the first N objects in the collection
        :param collection:
        :param limit:
        :param include:
        :param offset:
        :param kwargs:
        :return:
        """
        if include is None:
            include = {IDS, METADATAS, DOCUMENTS}
        else:
            include = set(include)
        safe_collection_name = f'"{collection}"'
        results = self.conn.execute(
            f"""
                SELECT id, metadata, embeddings, documents, NULL as distance
                FROM {safe_collection_name}
                LIMIT ?
            """,
            [limit],
        ).fetchall()

        yield from self.parse_duckdb_result(results, include)

    def fetch_all_objects_memory_safe(
        self, collection: str = None, batch_size: int = 100, include=None, **kwargs
    ) -> Iterator[OBJECT]:
        """
        Fetch all objects from a collection, in batches to avoid memory overload.
        """
        collection = self._get_collection(collection)
        offset = 0
        while True:
            if include is None:
                include = [IDS, METADATAS, DOCUMENTS, EMBEDDINGS]
            safe_collection_name = f'"{collection}"'
            query = f"""
                                SELECT id, metadata, embeddings, documents, NULL as distance
                                FROM {safe_collection_name}
                                LIMIT ? OFFSET ?
                            """
            results = self.conn.execute(query, [batch_size, offset]).fetchall()
            if results:
                yield from self.parse_duckdb_result(results, include)
                offset += batch_size
            else:
                break

    def get_raw_objects(self, collection) -> Iterator[Dict]:
        """
        Get all raw objects in the collection as they were inserted into the database
        :param collection:
        :return:
        """
        safe_collection_name = f'"{collection}"'
        results = self.conn.execute(
            f"""
                SELECT metadata
                FROM {safe_collection_name}
            """
        ).fetchall()
        for result in results:
            yield json.loads(result[0])

    def dump_then_load(
        self,
        collection: str = None,
        target: DBAdapter = None,
    ):
        """
        Dump the collection to a file and then load it into the target adapter
        :param collection:
        :param target:
        :param temp_file:
        :param format:
        :return:
        """
        if collection is None:
            raise ValueError("Collection name must be provided.")
        if not isinstance(target, DuckDBAdapter):
            raise ValueError("Target must be a DuckDBAdapter instance")

        result = self.get_raw_objects(collection)

        metadata = self.collection_metadata(collection)
        model = metadata["model"]
        vec_dimension = self._get_embedding_dimension(model)
        distance = metadata["hnsw_space"]
        # in case it exists already, remove
        target.remove_collection(collection, exists_ok=True)
        # using same collection name in target database
        target._create_table_if_not_exists(collection, vec_dimension, distance, model)
        target.set_collection_metadata(collection, metadata)
        batch_size = 5000
        for i in range(0, len(list(result)), batch_size):
            batch = result[i : i + batch_size]
            target.insert(batch, collection=collection)

    @staticmethod
    def kill_process(pid):
        """
        Kill the process with the given PID
        Returns
        -------

        """
        process = None
        try:
            process = psutil.Process(pid)
            process.terminate()  # Sends SIGTERM
            process.wait(timeout=5)
        except psutil.NoSuchProcess:
            logger.info("Process already terminated.")
        except psutil.TimeoutExpired:
            if process is not None:
                logger.warning("Process did not terminate in time, forcing kill.")
                process.kill()  # Sends SIGKILL as a last resort
        except Exception as e:
            logger.error(f"Failed to terminate process: {e}")

    @staticmethod
    def _generate_sql_command(collection: str, method: str) -> str:
        safe_collection_name = f'"{collection}"'
        if method == "insert":
            return f"""
                INSERT INTO {safe_collection_name} (id,metadata, embeddings, documents) VALUES (?, ?, ?, ?)
                """
        else:
            raise ValueError(f"Unknown method: {method}")

    def _is_openai(self, collection: str) -> bool:
        """
        Check if the collection uses a OpenAI Embedding model
        :param collection:
        :return:
        """
        collection = self._get_collection(collection)
        safe_collection_name = f'"{collection}"'
        query = f"SELECT metadata FROM {safe_collection_name} WHERE id = '__metadata__'"
        result = self.conn.execute(query).fetchone()
        if result:
            metadata = json.loads(result[0])
            if "model" in metadata and metadata["model"].startswith("openai:"):
                return True
        return False

    def _dict(self, obj: OBJECT):
        if isinstance(obj, dict):
            return obj
        elif isinstance(obj, BaseModel):
            return obj.model_dump(exclude_unset=True)
        elif isinstance(obj, YAMLRoot):
            return json_dumper.to_dict(obj)
        else:
            raise ValueError(f"Cannot convert {obj} to dict")

    def _id(self, obj, id_field):
        if isinstance(obj, dict):
            id = obj.get(id_field, None)
        else:
            id = getattr(obj, id_field, None)
        if not id:
            id = str(obj)
        self.id_to_object[id] = obj
        return id

    def _text(self, obj: OBJECT, text_field: Union[str, Callable]):
        if isinstance(obj, DuckDBSearchResult):
            obj = obj.dict()
        if isinstance(obj, list):
            raise ValueError(f"Cannot handle list of text fields: {obj}")
        if text_field is None or (isinstance(text_field, str) and text_field not in obj):
            obj = {k: v for k, v in obj.items() if v}
            t = yaml.safe_dump(obj, sort_keys=False)
        elif isinstance(text_field, Callable):
            t = text_field(obj)
        elif isinstance(obj, dict):
            t = obj[text_field]
        else:
            t = getattr(obj, text_field)
        t = t.strip()
        if not t:
            raise ValueError(f"Text field {text_field} is empty for {type(obj)} : {obj}")
        if len(t) > self.default_max_document_length:
            logger.warning(f"Truncating text field {text_field} for {str(obj)[0:100]}...")
            t = t[: self.default_max_document_length]
        return t

    def identifier_field(self, collection: str = None) -> str:
        if self.schema_proxy and self.schema_proxy.schemaview:
            fields = []
            for s in self.schema_proxy.schemaview.all_slots(attributes=True).values():
                if s.identifier:
                    fields.append(s.name)
            if fields:
                if len(fields) > 1:
                    raise ValueError(f"Multiple identifier fields: {fields}")
                return fields[0]
        return "id"

    @staticmethod
    def parse_duckdb_result(results, include) -> Iterator[SEARCH_RESULT]:
        """
        Parse the results from the SQL
        :return: DuckDBSearchResultIterator
        ----------
        """
        for res in results:
            if res[0] != "__metadata__":
                D = DuckDBSearchResult(
                    ids=res[0],
                    metadatas=json.loads(res[1]),
                    embeddings=res[2],
                    documents=res[3],
                    distances=res[4],
                    include=include,
                )
                yield from D.__iter__()

    @staticmethod
    def _parse_where_clause(where: Dict[str, Any]) -> str:
        """
        Parse the where clause from the query
        Parameters
        ----------
        where

        Returns
        -------

        """
        conditions = []
        for key, condition in where.items():
            if isinstance(condition, dict):
                for op, value in condition.items():
                    if op == "$eq":
                        conditions.append(f"json_extract_string(metadata, '$.{key}') = '{value}'")
                    elif op == "$ne":
                        conditions.append(f"json_extract_string(metadata, '$.{key}') != '{value}'")
                    elif op == "$gt":
                        conditions.append(
                            f"CAST(json_extract_string(metadata, '$.{key}') AS DOUBLE) > '{value}'"
                        )
                    elif op == "$gte":
                        conditions.append(
                            f"CAST(json_extract_string(metadata, '$.{key}') AS DOUBLE) >= '{value}'"
                        )
                    elif op == "$lt":
                        conditions.append(
                            f"CAST(json_extract_string(metadata, '$.{key}') AS DOUBLE) < '{value}'"
                        )
                    elif op == "$lte":
                        conditions.append(
                            f"CAST(json_extract_string(metadata, '$.{key}') AS DOUBLE) <= '{value}'"
                        )
                    elif op == "$in":
                        conditions.append(
                            f"json_extract_string(metadata, '$.{key}') IN ({', '.join([f'{v}' for v in value])})"
                        )
                    elif op == "$nin":
                        conditions.append(
                            f"json_extract_string(metadata, '$.{key}') NOT IN ({', '.join([f'{v}' for v in value])})"
                        )
                    elif op == "$exists":
                        if value:
                            conditions.append(f"json_extract(metadata, '$.{key}') IS NOT NULL")
                        else:
                            conditions.append(f"json_extract(metadata, '$.{key}') IS NULL")
                    elif op == "$regex":
                        conditions.append(f"json_extract_string(metadata, '$.{key}') ~ '{value}'")
            else:
                conditions.append(f"json_extract_string(metadata, '$.{key}') = '{condition}'")
        return " AND ".join(conditions)

    def _get_embedding_dimension(self, model_name: str) -> int:
        if model_name is None or model_name.startswith(self.default_model):
            return DEFAULT_MODEL[self.default_model]
        if isinstance(model_name, str):
            if model_name.startswith("openai:"):
                model_key = model_name.split("openai:", 1)[1]
                if model_key == "" or model_key not in MODEL_MAP.keys():
                    model_key = DEFAULT_OPENAI_MODEL
                model_info = MODEL_MAP.get(model_key, DEFAULT_OPENAI_MODEL)
                return model_info[1]
            else:
                return MODEL_MAP[DEFAULT_OPENAI_MODEL][1]

    @staticmethod
    def _validate_ef_construction(value: int) -> int:
        """
        The number of candidate vertices to consider during the construction of the index. A higher value will result
        in a more accurate index, but will also increase the time it takes to build the index.
        Parameters
        ----------
        value

        Returns
        -------

        """
        if not (10 <= value <= 200):
            raise ValueError("ef_construction must be between 10 and 200")
        return value

    @staticmethod
    def _validate_ef_search(value: int) -> int:
        """
        The number of candidate vertices to consider during the search phase of the index.
        A higher value will result in a more accurate index, but will also increase the time it takes to perform a search.
        Parameters
        ----------
        value

        Returns
        -------

        """
        if not (10 <= value <= 200):
            raise ValueError("ef_search must be between 10 and 200")
        return value

    @staticmethod
    def _validate_m(value: int) -> int:
        """
        The maximum number of neighbors to keep for each vertex in the graph.
        A higher value will result in a more accurate index, but will also increase the time it takes to build the index.
        Parameters
        ----------
        value

        Returns
        -------

        """
        if not (5 <= value <= 48):
            raise ValueError("M must be between 5 and 48")
        return value

    @staticmethod
    def determine_fields_to_include(include: Optional[List[str]] = None) -> str:
        """
        Determine which fields to include in the SQL query based on the 'include' parameter.

        :param include: List of fields to include in the output ['metadata', 'embeddings', 'documents']
        :return: Comma-separated string of fields to include
        """
        fields = []
        if include is None or IDS in include:
            fields.append(IDS)
        if include is None or METADATAS in include:
            fields.append(METADATAS)
        if include is None or EMBEDDINGS in include:
            fields.append(EMBEDDINGS)
        if include is None or DOCUMENTS in include:
            fields.append(DOCUMENTS)
        return ", ".join(fields)



"""Chat with a KB."""

import logging
import time
from dataclasses import dataclass
from typing import ClassVar, List

import inflection
import requests

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://en.wikipedia.org/w/api.php"


@dataclass
class WikipediaWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over Wikipedia.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "wikipedia"

    def external_search(self, text: str, expand: bool = False, **kwargs) -> List:
        # if expand:
        #    raise NotImplementedError
        # else:
        #    search_term = text
        search_term = text
        logger.info(f"Constructed search term: {search_term}")
        # Parameters for the request
        params = {"action": "query", "format": "json", "list": "search", "srsearch": search_term}

        time.sleep(0.5)
        response = requests.get(BASE_URL, params=params)
        data = response.json()
        search_results = data["query"]["search"]
        snippets = {result["title"]: result["snippet"] for result in search_results}
        titles = list(snippets.keys())

        info_params = {
            "action": "query",
            "format": "json",
            "titles": "|".join(titles),
            "prop": "extracts",
            "exintro": True,  # Get only the introduction of each page
            "explaintext": True,  # Get plain text instead of HTML
        }

        info_response = requests.get(BASE_URL, params=info_params)
        if not info_response.ok:
            raise ValueError(f"Could not get info for {titles}")
        info_data = info_response.json()

        # Extracting and printing the information
        if "query" not in info_data:
            logger.error(f"Could not get pages from {info_data}")
            return []
        pages = info_data["query"]["pages"]
        results = []
        for _page_id, page_info in pages.items():
            page_info["id"] = inflection.camelize(page_info["title"])
            page_info["snippet"] = snippets[page_info["title"]]
            del page_info["ns"]
            del page_info["pageid"]
            results.append(page_info)
        return results



from dataclasses import dataclass, field

from oaklib import get_adapter

from curategpt.wrappers.ontology.ontology_wrapper import OntologyWrapper


@dataclass
class BioportalWrapper(OntologyWrapper):
    """
    A wrapper over the Bioportal API.

    This makes use of OAK.
    """

    fetch_definitions: bool = field(default=False)
    fetch_relationships: bool = field(default=False)

    def __post_init__(self):
        if not self.oak_adapter:
            self.oak_adapter = get_adapter("bioportal:")
